### This script combines position data from multiple cameras.
### It also corrects frame time offset errors in PosLog.csv files
### It also removes bad position data lines

### Use as follows:
### import CombineTrackingData as combPos
### combPos.combdata(Path-To-Recording-Folder)

### By Sander Tanni, May 2017, UCL

from itertools import combinations

import numpy as np
from scipy.spatial.distance import euclidean
from openEPhys_DACQ import NWBio
from tqdm import tqdm


def combineCamerasData(cameraPos, lastCombPos, cameraIDs, CameraSettings, arena_size):
    # This outputs position data based on which camera is closest to tracking target.

    # cameraPos - list of numpy vecors with 4 elements (x1,y1,x2,y2) for each camera
    # lastCombPos - Last known output from this function. If None, the function will attempt to locate the animal.
    # cameraIDs - list of of CameraSettings.keys() in corresponding order to cameraPos and lastCombPos
    # CameraSettings - settings dictionary created by CameraSettings.CameraSettingsApp
    # arena_size - 2 element numpy array with arena height and width.

    # Output - numpy vector (x1,y1,x2,y2) with data from closest camera

    # Animal detection finding method in case lastCombPos=None
    #   simultaneously from at least 2 cameras with smaller separation than half of CameraSettings['camera_transfer_radius']
    #   If successful, closest mean coordinate is set as output
    #   If unsuccessful, output is None

    N_RPis = len(cameraPos)
    cameraPos = np.array(cameraPos, dtype=np.float32)
    camera_locations = []
    for cameraID in cameraIDs:
        camera_locations.append(CameraSettings['CameraSpecific'][cameraID]['location_xy'])
    camera_locations = np.array(camera_locations, dtype=np.float32)

    # Only work with camera data from inside the enviornment
    # Find bad pos data lines
    idxBad = np.zeros(cameraPos.shape[0], dtype=bool)
    # Points beyond arena size
    x_too_big = cameraPos[:,0] > arena_size[0] + 20
    y_too_big = cameraPos[:,1] > arena_size[1] + 20
    idxBad = np.logical_or(idxBad, np.logical_or(x_too_big, y_too_big))
    x_too_small = cameraPos[:,0] < -20
    y_too_small = cameraPos[:,1] < -20
    idxBad = np.logical_or(idxBad, np.logical_or(x_too_small, y_too_small))
    # Only keep camera data from within the environment
    N_RPis = np.sum(np.logical_not(idxBad))
    # Only continue if at least one RPi data remains
    if N_RPis > 0:
        cameraPos = cameraPos[np.logical_not(idxBad),:]
        camera_locations = camera_locations[np.logical_not(idxBad),:]
        if np.any(lastCombPos):
            # Check which cameras provide data close enough to lastCombPos
            RPi_correct = []
            for nRPi in range(N_RPis):
                lastCombPos_distance = euclidean(cameraPos[nRPi, :2], lastCombPos[:2])
                RPi_correct.append(lastCombPos_distance < CameraSettings['General']['camera_transfer_radius'])
            RPi_correct = np.array(RPi_correct, dtype=bool)
            # If none were found to be withing search radius, set output to None
            if not np.any(RPi_correct):
                combPos = None
            else:
                # Use the reading from closest camera to target mean location that detects correct location
                if np.sum(RPi_correct) > 1:
                    # Only use correct cameras
                    N_RPis = np.sum(RPi_correct)
                    cameraPos = cameraPos[RPi_correct, :]
                    camera_locations = camera_locations[RPi_correct, :]
                    meanPos = np.mean(cameraPos[:, :2], axis=0)
                    # Find mean position distance from all cameras
                    cam_distances = []
                    for nRPi in range(N_RPis):
                        camera_loc = camera_locations[nRPi, :]
                        cam_distances.append(euclidean(camera_loc, meanPos))
                    # Find closest distance camera and output its location coordinates
                    closest_camera = np.argmin(np.array(cam_distances))
                    combPos = cameraPos[closest_camera, :]
                else:
                    # If target only detected close enough to lastCombPos in a single camera, use it as output
                    combPos = cameraPos[np.where(RPi_correct)[0][0], :]
        else:
            # If no lastCombPos provided, check if position can be verified from more than one camera
            #### NOTE! This solution breaks down if more than two cameras incorrectly identify the same object
            ####       as the brightes spot, instead of the target LED.
            cameraPairs = []
            pairDistances = []
            for c in combinations(range(N_RPis), 2):
                pairDistances.append(euclidean(cameraPos[c[0], :2], cameraPos[c[1], :2]))
                cameraPairs.append(np.array(c))
            cameraPairs = np.array(cameraPairs)
            cameraPairs_Match = np.array(pairDistances) < (CameraSettings['General']['camera_transfer_radius'] / 2.0)
            # If position can not be verified from multiple cameras, set output to none
            if not np.any(cameraPairs_Match):
                combPos = None
            else:
                # Otherwise, set output to mean of two cameras with best matching detected locations
                pairToUse = np.argmin(pairDistances)
                camerasToUse = np.array(cameraPairs[pairToUse, :])
                combPos = np.mean(cameraPos[camerasToUse, :2], axis=0)
                # Add NaN values for second LED
                combPos = np.append(combPos, np.empty(2) * np.nan)
    else:
        combPos = None

    return combPos


def remove_tracking_data_outside_boundaries(posdata, arena_size, max_error=20):
    NotNaN = np.where(np.logical_not(np.isnan(posdata[:,1])))[0]
    idxBad = np.zeros(NotNaN.size, dtype=bool)
    x_too_big = posdata[NotNaN,1] > arena_size[0] + max_error
    y_too_big = posdata[NotNaN,2] > arena_size[1] + max_error
    idxBad = np.logical_or(idxBad, np.logical_or(x_too_big, y_too_big))
    x_too_small = posdata[NotNaN,1] < -max_error
    y_too_small = posdata[NotNaN,2] < -max_error
    idxBad = np.logical_or(idxBad, np.logical_or(x_too_small, y_too_small))
    # Combine good position data lines
    posdata = np.delete(posdata, NotNaN[idxBad], axis=0)

    return posdata


def remove_tracking_data_jumps(posdata, maxjump):
    """
    Removes data with too large jumps based on euclidean distance

    posdata - numpy array with columns:
              timestamps
              LED 1 xpos
              LED 1 ypos
              LED 2 xpos
              LED 2 ypos
              , where NaN for missing LED 2 data
    maxjump - int or float specifying maximum allowed shift in euclidean distance
    """
    keepPos = []
    lastPos = posdata[0,1:3]
    for npos in range(posdata.shape[0]):
        currpos = posdata[npos,1:3]
        if euclidean(lastPos, currpos) < maxjump:
            keepPos.append(npos)
            lastPos = currpos
    keepPos = np.array(keepPos)
    print(str(posdata.shape[0] - keepPos.size) + ' of ' + 
          str(posdata.shape[0]) + ' removed in postprocessing')
    posdata = posdata[keepPos,:]

    return posdata


def iteratively_combine_multicamera_data_for_recording(
    CameraSettings, arena_size, posdatas, OE_GC_times, verbose=False):
    """Return ProcessedPos

    Combines raw tracking data from multiple cameras into a single ProcessedPos array.
    If a single camera data is provided, this will be converted into same format.
    """
    cameraIDs = sorted(CameraSettings['CameraSpecific'].keys())
    # Load position data for all cameras
    for i, posdata in enumerate(posdatas):
        if isinstance(posdata, dict):
            # This is conditional because for early recordings tracking data was immediately stored in Open Ephys time.
            # Remove datapoints where all position data is None
            idx_None = np.all(np.isnan(posdata['OnlineTrackerData']), 1)
            posdata['OnlineTrackerData'] = np.delete(posdata['OnlineTrackerData'], 
                                                     np.where(idx_None)[0], axis=0)
            posdata['OnlineTrackerData_timestamps'] = np.delete(posdata['OnlineTrackerData_timestamps'], 
                                                                np.where(idx_None)[0], axis=0)
            # Compute position data timestamps in OpenEphys time
            RPi_frame_in_OE_times = NWBio.estimate_open_ephys_timestamps_from_other_timestamps(
                OE_GC_times,
                posdata['GlobalClock_timestamps'],
                posdata['OnlineTrackerData_timestamps'],
                other_times_divider=10 ** 6
            )

            # Combine timestamps and position data into a single array
            posdata = np.concatenate((RPi_frame_in_OE_times.astype(np.float64)[:, None], posdata['OnlineTrackerData']), axis=1)
            posdatas[i] = posdata
    if len(posdatas) > 1:
        # If data from multiple cameras available, combine it
        PosDataFramesPerSecond = 30.0
        # Find first and last timepoint for position data
        first_timepoints = []
        last_timepoints = []
        for posdata in posdatas:
            first_timepoints.append(posdata[0, 0])
            last_timepoints.append(posdata[-1, 0])
        # Combine position data step-wise from first to last timepoint at PosDataFramesPerSecond
        # At each timepoint the closest matchin datapoints will be taken from different cameras
        timepoints = np.arange(np.array(first_timepoints).min(), np.array(last_timepoints).max(), 1.0 / PosDataFramesPerSecond)
        combPosData = [None]
        listNaNs = []
        if verbose:
            print('Combining iteratively camera data for each position timepoint')
        for npoint in (tqdm(range(len(timepoints))) if verbose else range(len(timepoints))):
            # Find closest matchin timepoint from all RPis
            idx_tp = np.zeros(4, dtype=np.int32)
            for nRPi in range(len(posdatas)):
                idx_tp[nRPi] = np.argmin(np.abs(timepoints[npoint] - posdatas[nRPi][:,0]))
            # Convert posdatas for use in combineCamerasData function
            cameraPos = []
            for nRPi in range(len(posdatas)):
                cameraPos.append(posdatas[nRPi][idx_tp[nRPi], 1:5])
            tmp_comb_data = combineCamerasData(cameraPos, combPosData[-1], cameraIDs, CameraSettings, arena_size)
            combPosData.append(tmp_comb_data)
            if tmp_comb_data is None:
                listNaNs.append(npoint)
        # Remove the extra element from combPosData
        del combPosData[0]
        # Remove all None elements
        for nanElement in listNaNs[::-1]:
            del combPosData[nanElement]
        timepoints = np.delete(timepoints, listNaNs)
        # Combine timepoints and position data
        ProcessedPos = np.concatenate((np.expand_dims(np.array(timepoints), axis=1), np.array(combPosData)), axis=1)
        # Print info about None elements
        if len(listNaNs) > 0:
            print('Total of ' + str(len(listNaNs) * (1.0 / PosDataFramesPerSecond)) + ' seconds of position data was lost')
            if listNaNs[0] == 0:
                print('This was in the beginning and ' + str(np.sum(np.diff(np.array(listNaNs)) > 1)) + ' other epochs.')
            else:
                print('This was not in the beginning, but in ' + str(np.sum(np.diff(np.array(listNaNs)) > 1) + 1) + ' other epochs.')
    else:
        # In case of single camera being used, just use data from that camera
        ProcessedPos = posdatas[0]
    ProcessedPos = remove_tracking_data_outside_boundaries(ProcessedPos, arena_size, max_error=10)
    ProcessedPos = ProcessedPos.astype(np.float64)

    return ProcessedPos




import numpy as np
from scipy.spatial.distance import euclidean
from itertools import combinations
import cv2

from openEPhys_DACQ import NWBio
from openEPhys_DACQ.video_io import RecordingCameraVideo


def transform_pix_to_real_value(y_ind, x_ind, calibration_matrix):
    """ Transforms position on image from pixels to centimeters in real values using a transformation matrix.

    :param float y_ind: y-axis (first/vertical dimension) position on image
    :param float x_ind: x-axis (second/horizontal dimension) position on image
    :param calibration_matrix:
    :return: x_val, y_val - corresponding real values to x_ind and y_ind
    """
    led_pix_4pt = np.reshape(np.array([y_ind, x_ind]), (1, 1, 2)).astype(np.float32)
    pos_cm_from4pt = cv2.perspectiveTransform(led_pix_4pt, calibration_matrix).astype('float')
    x_val, y_val = pos_cm_from4pt.squeeze()

    return x_val, y_val


def image_peak_ind(image):
    """Returns peak y (horizontal) and x (vertical) indices for grayscale image

    :param numpy.ndarray image: grayscale image of shape (M, N)
    :return: y_ind, x_ind
    """
    (_1, _2, _3, (x_ind, y_ind)) = cv2.minMaxLoc(image)
    return y_ind, x_ind


def crop_image_around_point(image, y_ind, x_ind, min_radius):
    """Returns a rectangular crop of the image respecting image boundaries, such that it can fit a circle of min_radius.

    If y_ind and x_ind are too close to the edge, the cropped image will be smaller in the direction of the edge(s).

    :param numpy.ndarray image: grayscale image shape (M, N)
    :param int y_ind: center of cropped image relative to input image along y (vertical/first) dimension
    :param int x_ind: center of cropped image relative to input image along x (horizontal/second) dimension
    :param int min_radius: cropped image size is such that it would perfectly fit a circle with min_radius radius.
    :return: cropped_image, y_min, x_min
    """

    x_min = x_ind - min_radius
    x_max = x_ind + min_radius
    y_min = y_ind - min_radius
    y_max = y_ind + min_radius

    x_min = x_min if x_min > 0 else 0
    y_min = y_min if y_min > 0 else 0
    x_max = x_max if x_max <= image.shape[1] else image.shape[1]
    y_max = y_max if y_max <= image.shape[0] else image.shape[0]

    return image[x_min:y_max, x_min:x_max], y_min, x_min


def find_bright_circular_blobs(image, threshold):
    """Returns list of blobs in order of descending size and keypoints element for drawing blobs with opencv.

    :param numpy.ndarray image: grayscale image shape (M, N)
    :param float threshold: threshold value for blob detection
    :return: blobs, keypoints
    """
    params = cv2.SimpleBlobDetector_Params()
    params.minThreshold = threshold
    params.thresholdStep = 0.01 * threshold
    params.filterByArea = True
    params.minArea = 2
    params.filterByCircularity = True
    params.minCircularity = 0.95
    params.filterByConvexity = True
    params.minConvexity = 0.95
    params.minDistBetweenBlobs = 2
    params.blobColor = 255

    detector = cv2.SimpleBlobDetector_create(params)

    keypoints = detector.detect(image)

    blobs = [{'size': keypoint.size, 'x_loc': keypoint.pt[0], 'y_loc': keypoint.pt[1]} for keypoint in keypoints]

    _, blobs, keypoints = sorted(zip([blob['size'] for blob in blobs], blobs, keypoints), reverse=True)

    return blobs, keypoints


def shift_cropped_image_blobs(blobs, keypoints, y_min, x_min):
    """Shift positional values of output from :py:func:`find_bright_circular_blobs` by x_min and y_min.
    """
    for blob in blobs:
        blob['x_loc'] = blob['x_loc'] + x_min
        blob['y_loc'] = blob['y_loc'] + y_min
    for keypoint in keypoints:
        keypoint.pt = (keypoint.pt[0] + x_min, keypoint.pt[1] + y_min)


def distance_between_adjacent_pixels(calibration_matrix, frame_shape):
    """Approximation of distance in centimeters correspond to single pixel difference at the center of field of view.

    :param numpy.ndarray calibration_matrix: matrix that can be used with :py:func:`cv2.perspectiveTransform`
    :param tuple frame_shape: shape of the frames used
    :return: distance
    """
    distance_in_pixels = 10
    # Pick two points at the center of the field of view
    tmp_loc1 = np.reshape(np.array([int(frame_shape[1] / 2), int(frame_shape[0] / 2)], dtype=np.float32), (1, 1, 2))
    tmp_loc2 = \
        np.reshape(np.array([int(frame_shape[1] / 2), int(frame_shape[0] / 2) + distance_in_pixels], dtype=np.float32),
                   (1, 1, 2))
    # Use transformation matrix to map pixel values to position in real world in centimeters
    tmp_loc1 = cv2.perspectiveTransform(tmp_loc1, calibration_matrix)
    tmp_loc2 = cv2.perspectiveTransform(tmp_loc2, calibration_matrix)
    # Compute the distance between the two points in real world centimeters
    real_distance = euclidean(np.array([tmp_loc1[0, 0, 0].astype('float'), tmp_loc1[0, 0, 1].astype('float')]),
                              np.array([tmp_loc2[0, 0, 0].astype('float'), tmp_loc2[0, 0, 1].astype('float')]))
    # Compute distance in centimeters between adjacent pixels
    distance = float(real_distance) / float(distance_in_pixels)

    return distance


def convert_centimeters_to_pixel_distance(real_distance, calibration_matrix, frame_shape):
    """Converts a distance in real values to distance in pixels at the center of frame.

    :param float real_distance: distance in real values
    :param numpy.ndarray calibration_matrix: matrix that can be used with :py:func:`cv2.perspectiveTransform`
    :param tuple frame_shape: shape of the frames used
    :return: pixel_distance
    """
    return int(np.round(float(real_distance) / distance_between_adjacent_pixels(calibration_matrix, frame_shape)))


class LedDetector(object):
    """
    Detects blobs around brightest position in images.
    """

    def __init__(self, image_shape, calibration_matrix, detection_window, search_radius,
                 threshold_multiplier=0.75, smoothing_size=5):
        """
        :param tuple image_shape: (N, M) specifying shape of images used
        :param numpy.ndarray calibration_matrix: matrix that can be used with :py:func:`cv2.perspectiveTransform`
        :param detection_window: (x_min, x_max, y_min, y_max) range of real world values where to look for led.
        :param float search_radius: size of search area (in real values) around brightest spot for blobs.
        :param float threshold_multiplier: used to multiply value at brightest spot for blob detection threshold.
        :param float smoothing_size: sigma of spatial smoothing gaussian kernel.
        """
        self._image_shape = image_shape
        self._calibration_matrix = calibration_matrix
        self._search_radius = convert_centimeters_to_pixel_distance(search_radius, calibration_matrix, image_shape)
        self._threshold_multiplier = threshold_multiplier
        self._smoothing_size = smoothing_size

        # Compute position of mapping of each pixel to real value based on calibration_matrix
        self._real_value_map = np.array((image_shape[0], image_shape[1], 2), dtype=np.float32)
        for y_ind in range(image_shape[0]):
            for x_ind in range(image_shape[1]):
                self._real_value_map[y_ind, x_ind, :] = transform_pix_to_real_value(y_ind, x_ind, calibration_matrix)

        # Create mask for masking out image pixels mapping to positions outside detection_window
        self._mask = np.zeros(image_shape, dtype=np.bool)
        for y_ind in range(image_shape[0]):
            for x_ind in range(image_shape[1]):
                if (detection_window[0] <= x_ind <= detection_window[1]
                        and detection_window[2] <= y_ind <= detection_window[3]):
                    self._mask[y_ind, x_ind] = False

    def process(self, image):
        """Returns data on bright blobs found around brightest point in sorted order of largest blobs first.

        blobs - list of dicts with elements:
            'x_loc' - x axis (horizontal/second dimension) image pixel coordinate of blob
            'y_loc' - y axis (vertical/first dimension) image pixel coordinate of blob
            'x_real' - x axis (horizontal/second dimension) real coordinate of blob based on calibration_matrix
            'y_real' - y axis (vertical/first dimension) real coordinate of blob based on calibration_matrix
            'size' - size of the blob

        keypoints - raw output from :py:func:`cv2.SimpleBlobDetector_create.detect` that can be used for
            visualising blobs with:
        >>> cv2.drawKeypoints(image, keypoints, np.array([]), (0, 0, 255),
        >>>                   cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS)

        :param numpy.ndarray image: grayscale image with shape matching image_shape given during initialization
            gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        :return: blobs, keypoints
        """
        if image.shape != self._image_shape:
            raise Exception('Image shape does not match image_shape specified during initiation.')

        image = cv2.GaussianBlur(image, (0, 0), self._smoothing_size)
        image[self._mask] = 0

        y_ind, x_ind = image_peak_ind(image)

        cropped_image, y_min, x_min = crop_image_around_point(image, y_ind, x_ind, self._search_radius)
        blobs, keypoints = find_bright_circular_blobs(cropped_image, image[y_ind, x_ind] * self._threshold_multiplier)
        shift_cropped_image_blobs(blobs, keypoints, y_min, x_min)

        for blob in blobs:
            x, y = transform_pix_to_real_value(blob['y_loc'], blob['x_loc'], self._calibration_matrix)
            blob['x_real'] = x
            blob['y_real'] = y

        return blobs, keypoints


class LimitedMemory(object):

    def __init__(self, max_length):

        self._max_length = max_length
        self._items = []
        self._identifiers = []

    def __contains__(self, identifier):
        return identifier in self._identifiers

    def __len__(self):
        return len(self._items)

    def __getitem__(self, key):
        return self._items[key]

    @property
    def items(self):
        return self._items

    @property
    def identifiers(self):
        return self._identifiers

    def append(self, other, identifier=None):
        if len(self._items) >= self._max_length:
            self._items.pop(0)
            self._identifiers.pop(0)
        self._items.append(other)
        self._identifiers.append(identifier)

    def get(self, identifier):
        if identifier in self._identifiers:
            return self._items[self._identifiers.index(identifier)]
        else:
            raise ValueError('identifier {} not found in {}'.format(identifier, self))


def identify_true_position(led_positions, max_separation):
    """Returns x and y coordinates of true led_position identified and None otherwise.
    
    :param dict led_positions: dictionary with camera_id as keys and elements as dictionaries of led_position,
        as output from :py:func:`LedDetector.process`. This method uses the 'y_real' and 'x_real' values.
    :param float max_separation: maximum distance allowed between led_position from two cameras for confirmation.
    :return: x, y or None
    """

    # Compute distances between positions of all camera pairs
    camera_pairs = []
    camera_pair_distances = []
    for camera_ids in combinations(list(map(str, led_positions.keys())), 2):

        if len(led_positions[camera_ids[0]]) == 0:
            continue

        camera_pair_distances.append(euclidean(
            np.array((led_positions[camera_ids[0]][0]['x_real'], led_positions[camera_ids[0]][0]['y_real'])),
            np.array((led_positions[camera_ids[1]][0]['x_real'], led_positions[camera_ids[1]][0]['y_real']))
        ))
        camera_pairs.append(camera_ids)

    camera_pair_close_enough_idx = np.array(camera_pair_distances) < max_separation

    if not np.any(camera_pair_close_enough_idx):

        return None

    else:

        closest_camera_pair_ind = int(np.argmin(camera_pair_distances))
        camera_ids = camera_pairs[closest_camera_pair_ind]

        x, y = np.mean(np.array([(led_positions[camera_ids[0]][0]['x_real'],
                                  led_positions[camera_ids[0]][0]['y_real']),
                                 (led_positions[camera_ids[1]][0]['x_real'],
                                  led_positions[camera_ids[1]][0]['y_real'])]),
                       axis=0)

        return x, y


class ProcessedLedLocationState(object):

    unknown_position = np.array((np.nan, np.nan))

    def __init__(self):
        self._last_confirmed_position = self.unknown_position
        self._last_confirmed_timestamp = None
        self._last_timestamp_was_confirmed = False

    @property
    def last_confirmed_position(self):
        return self._last_confirmed_position

    @property
    def last_confirmed_timestamp(self):
        return self._last_confirmed_timestamp

    @property
    def last_timestamp_was_confirmed(self):
        return self._last_timestamp_was_confirmed

    def _update_prediction(self, new_position):
        if np.any(np.isnan(new_position)) or np.any(np.isnan(self._last_confirmed_position)):
            self.predicted_position = self.unknown_position
        else:
            last_movement_vector = new_position - self._last_confirmed_position
            self.predicted_position = new_position + last_movement_vector

    def update(self, position, timestamp):
        position = np.array(position)
        self._update_prediction(position)
        if ~np.any(np.isnan(position)):
            self._last_confirmed_position = position
            self._last_confirmed_timestamp = timestamp
            self._last_timestamp_was_confirmed = True
        else:
            self._last_timestamp_was_confirmed = False


class DualLedMultiCameraTracker(object):

    def __init__(self, led_1_xy, led_2_xy, timestamp, camera_positions,
                 flip_reset_speed=40, flip_reset_steps=6, flip_reset_tortuosity=0.8):

        self._led_states = (ProcessedLedLocationState(),
                            ProcessedLedLocationState())
        self._led_states[0].update(led_1_xy, timestamp)
        self._led_states[1].update(led_2_xy, timestamp)

        self._camera_positions = camera_positions

        self._flip_reset_speed = flip_reset_speed
        self._flip_reset_tortuosity = flip_reset_tortuosity

        self._position_history = LimitedMemory(flip_reset_steps)

        self._anterior_led_ind = 0
        self._posterior_led_ind = 1

    @staticmethod
    def position_vector_angle(position_vector):
        return np.angle(complex(*position_vector)) + np.pi

    def _get_movement_direction_for_led_flip(self):

        positions = np.array(self._position_history.items)

        if np.any(np.isnan(positions).flatten()):
            return None

        timestamps = np.array(self._position_history.identifiers)

        position_deltas = np.sqrt(np.sum(np.diff(positions, axis=0) ** 2, axis=1))

        if np.sum(position_deltas) / euclidean(positions[0, :], positions[-1, :]) < self._flip_reset_tortuosity:
            return None

        if np.mean(position_deltas / np.diff(timestamps)) < self._flip_reset_speed:
            return None

        full_movement_vector = positions[-1, :] - positions[0, :]

        return self.position_vector_angle(full_movement_vector)

    def _attempt_led_flip(self):
        """Checks self._led_state values and compares to recent trajectory direction. If they are in
        opposite directions, the led identities are flipped.
        """

        if not all([led_state.last_timestamp_was_confirmed for led_state in self._led_states]):
            return

        movement_direction = self._get_movement_direction_for_led_flip()

        if movement_direction is None:
            return

        led_direction = self.position_vector_angle(self._led_states[self._anterior_led_ind]
                                                   - self._led_states[self._posterior_led_ind])

        if np.mod(movement_direction - led_direction, 2 * np.pi) > np.pi:

            self._anterior_led_ind = (1, 0)[self._anterior_led_ind]
            self._posterior_led_ind = (1, 0)[self._anterior_led_ind]

    def update(self, led_positions):

        # This method should use last known location to decide which camera position to use,
        # by comparing the distances to each of the cameras

        # There could be some inertia to staying with one camera beyond the mid-line between cameras.

        # It should match the input LED locations to the elements of self._led_states based
        # on proximity to the predicted locations.

        # The following function could also be used to further verify LED angle with respect to movement direction.
        self._attempt_led_flip()

        # It should then assign the led locations to correct elements in self._led_states based on
        # self._anterior_led_ind and self._posterior_led_ind.

        # If a camera switch happens, the self._anterior_led_ind and self._posterior_led_ind should default


class DualLedMultiCameraProcessor(object):
    """
    Tracks LED position in video data.
    """

    def __init__(self, video, calibration_matrix, camera_position, arena_size, led_separation,
                 threshold_multiplier, smoothing_size, camera_transfer_radius=None, init_camera_id=None):

        self._camera_ids = list(map(str, self._video.keys()))

        self._video = video
        self._calibration_matrix = calibration_matrix
        self._camera_position = camera_position

        if camera_transfer_radius is None and init_camera_id is None:
            raise ValueError('Either camera_transfer_radius or init_camera_id must be specified.')

        self._init_camera_id = init_camera_id
        self._camera_transfer_radius = camera_transfer_radius if self._init_camera_id is None else None

        self._led_detector = {}
        for camera_id in self._video:
            self._led_detector[camera_id] = LedDetector(
                self._video[camera_id].current_frame.shape,
                self._calibration_matrix[camera_id],
                (0, arena_size[0], 0, arena_size[1]),
                led_separation * 2,
                threshold_multiplier,
                smoothing_size
            )

        self._led_position_memory = {camera_id: LimitedMemory(10) for camera_id in self._video}

    def _get_led_position_in_current_frame(self, camera_id):
        """Returns led_position in current image of this camera.

        :param str camera_id: camera identifier
        :return: led_position
        """
        # If possible, use pre-existing LED positions
        if self._video[camera_id].current_timestamp in self._led_position_memory[camera_id]:
            return self._led_position_memory[camera_id].get(self._video[camera_id].current_timestamp)
        # Find LED positions and store in memory
        led_position = self._led_detector[camera_id].process(self._video[camera_id].current_frame)
        self._led_position_memory[camera_id].append(led_position, self._video[camera_id].current_timestamp)

        return led_position[:2] if len(led_position) > 2 else led_position

    def _attempt_to_identify_true_position(self):

        if len(self._video) == 1:

            led_position = self._get_led_position_in_current_frame(self._camera_ids[0])[0]

            return np.array(led_position['x_real'], led_position['y_real']), np.array((np.nan, np.nan))

        elif not (self._init_camera_id is None):

            led_position = self._get_led_position_in_current_frame(self._init_camera_id)[0]

            return np.array(led_position['x_real'], led_position['y_real']), np.array((np.nan, np.nan))

        else:

            led_positions = {camera_id: self._get_led_position_in_current_frame(camera_id)
                             for camera_id in self._camera_ids}

            xy = identify_true_position(led_positions, self._camera_transfer_radius / 2.)

            if xy is None:

                return np.array((np.nan, np.nan, np.nan, np.nan))

            else:

                return np.array(xy[0], xy[1]), np.array((np.nan, np.nan))

    def get_current_led_positions(self):
        # This method should be using class DualLedMultiCameraTracker to keep track of LED location.
        # DualLedMultiCameraTracker should notify if LED was not been tracked for a long period,
        # which should prompt this DualLedMultiCameraProcessor class to reinitialize tracking
        # using multi-camera confirmation method.


class OfflineTracker(object):
    """
    Uses video data stored on the disk to provide tracking information for a recording.
    """

    # List of supported processing methods in _initialize_processing_method and _process_for_current_timestamp
    supported_methods = ('dual_led',)

    def __init__(self, fpath, fps=30, method=None, threshold_multiplier=0.75):

        # Parse input

        self._fpath = NWBio.get_filename(fpath)

        if not NWBio.check_if_open_ephys_nwb_file(self._fpath):
            raise ValueError('Path {} does not lead to a recognised recording file.'.format(self._fpath))

        self._timestep = 1 / float(fps)

        settings = NWBio.load_settings(self._fpath)

        self._method = settings['CameraSettings']['General']['tracking_mode'] if method is None else method
        if not (self._method in self.supported_methods):
            raise ValueError('Method {} not found in support methods {}'.format(self._method, self.supported_methods))

        self._threshold_multiplier = threshold_multiplier

        # Get list of cameras_ids

        self._camera_ids = tuple(map(str, settings['CameraSettings']['CameraSpecific'].keys()))
        if len(self._camera_ids) == 0:
            raise Exception('No cameras specified in CameraSettings.')

        # Get settings for processing camera images

        resolution_setting = settings['CameraSettings']['General']['resolution_option']

        self._arena_size = settings['General']['arena_size']
        self._led_separation = settings['CameraSettings']['General']['LED_separation']
        self._camera_transfer_radius = settings['CameraSettings']['General']['camera_transfer_radius']
        self._smoothing_size = {'high': 4, 'low': 2}[resolution_setting]

        self._calibration_matrix = {}
        self._camera_position = {}

        for camera_id in self._camera_ids:
            camera_id_settings = settings['CameraSettings']['CameraSpecific'][camera_id]

            self._calibration_matrix[camera_id] = \
                camera_id_settings['CalibrationData'][resolution_setting]['calibrationTmatrix']
            self._camera_position[camera_id] = camera_id_settings['position_xy']

        # Initialise video feed

        self._video = {camera_id: RecordingCameraVideo(self._fpath, camera_id) for camera_id in self._camera_ids}

        self._align_videos()

        # Initialise tracking process

        self._current_timestamp = min([self._video[camera_id].current_timestamp
                                       for camera_id in self._camera_ids])

        self._final_timestamp = min([self._video[camera_id].final_timestamp
                                     for camera_id in self._camera_ids])

        self._timestamps = []
        self._tracking_data = []

        self._process_started = False
        self._process_finished = False

    def _align_videos(self):
        """Ensures first frames from all cameras are as closely aligned as possible.
        """
        if len(self._camera_ids) == 1:
            return

        frame_received = {camera_id: False for camera_id in self._camera_ids}
        while not all(frame_received.values()):

            earliest_frame_camera_id = self._camera_ids[int(np.argmin([self._video[camera_id].current_timestamp
                                                                       for camera_id in self._camera_ids]))]
            self._video[earliest_frame_camera_id].next()
            frame_received[earliest_frame_camera_id] = True

    def _seek_closest_frames_to_current_timestamp(self):
        """Moves camera to next video frame if it is closer to current timestamp.
        """
        camera_aligned = {camera_id: False for camera_id in self._camera_ids}
        while not all(camera_aligned.values()):

            for camera_id in self._camera_ids:

                if camera_aligned[camera_id]:
                    continue

                if self._video[camera_id].next_timestamp is None:
                    camera_aligned[camera_id] = True
                    continue

                current_frame_delta = abs(self._current_timestamp - self._video[camera_id].current_timestamp)
                next_frame_delta = abs(self._current_timestamp - self._video[camera_id].next_timestamp)

                if next_frame_delta < current_frame_delta:
                    ret = self._video[camera_id].next()
                    assert ret, ('This should be True as loop execution should not make it here \n'
                                 + 'if next frame does not exist. The video data may be corrupted.')
                else:
                    camera_aligned[camera_id] = True

    def _process_current_timestamp_with_dual_led_multi_camera_processor(self):
        pass

    def _initialize_processing_method(self):

        if self._method == 'dual_led':

            if len(self._camera_ids) > 1:

                self._processor = DualLedMultiCameraProcessor(self._video, self._calibration_matrix,
                                                              self._camera_position, self._arena_size,
                                                              self._led_separation, self._threshold_multiplier,
                                                              self._smoothing_size,
                                                              camera_transfer_radius=self._camera_transfer_radius)
                return

        raise Exception('No processor available to match method {} and n = {} camera(s).'.format(
            self._method, len(self._camera_ids)))

    def _process_for_current_timestamp(self):

        if self._method == 'dual_led':

            if len(self._camera_ids) > 1:

                return self._process_current_timestamp_with_dual_led_multi_camera_processor()

    def process(self):

        if self._process_started:
            raise Exception('Method process can be call only once for each instance of OfflineTracker.')
        else:
            self._process_started = True

        self._initialize_processing_method()

        while self._current_timestamp < self._final_timestamp:

            self._timestamps.append(self._current_timestamp)

            self._seek_closest_frames_to_current_timestamp()

            self._tracking_data.append(self._process_for_current_timestamp())

            self._current_timestamp += self._timestep

        self._process_finished = True

    def ensure_process_is_finished(self):
        if not self._process_finished:
            self.process()

    @property
    def timestamps(self):
        self.ensure_process_is_finished()
        return self._timestamps

    @property
    def led_positions(self):
        self.ensure_process_is_finished()
        return self._tracking_data

    def close(self):
        for camera_id in self._camera_ids:
            self._video[camera_id].close()



# -*- coding: utf-8 -*-

import h5py
import numpy as np
import os
import sys
from openEPhys_DACQ.HelperFunctions import tetrode_channels, channels_tetrode, closest_argmin
from pprint import pprint
from copy import copy
import argparse
import importlib
from tqdm import tqdm


def OpenEphys_SamplingRate():
    return 30000


def bitVolts():
    return 0.195


def spike_waveform_leftwards_shift():
    """Returns the leftwards shift of waveforms from detection point in seconds.

    :return:
    :rtype: float
    """
    return 6 * (1.0 / OpenEphys_SamplingRate())


def get_filename(path):
    if not os.path.isfile(path):
        return os.path.join(path, 'experiment_1.nwb')
    else:
        return path


def delete_path_in_file(filename, path):
    with h5py.File(filename, 'r+') as h5file:
        del h5file[path]


def get_recordingKey(filename):
    with h5py.File(filename, 'r') as h5file:
        return list(h5file['acquisition']['timeseries'].keys())[0]


def get_all_processorKeys(filename):
    with h5py.File(filename, 'r') as h5file:
        return list(h5file['acquisition']['timeseries'][get_recordingKey(filename)]['continuous'].keys())


def get_processorKey(filename):
    return get_all_processorKeys(filename)[0]


def get_all_processor_paths(filename):
    return ['/acquisition/timeseries/' + get_recordingKey(filename)
            + '/continuous/' + processorKey
            for processorKey in get_all_processorKeys(filename)]


def get_processor_path(filename):
    return '/acquisition/timeseries/' + get_recordingKey(filename) \
           + '/continuous/' + get_processorKey(filename)


def check_if_open_ephys_nwb_file(filename):
    """
    Returns True if processor path can be identified
    in the file and False otherwise.
    """
    try:
        processor_path = get_processor_path(filename)
        with h5py.File(filename, 'r') as h5file:
            return processor_path in h5file
    except:
        return False


def get_downsampled_data_paths(filename):
    """
    Returns paths to downsampled data in NWB file.

    :param filename: path to NWB file
    :type filename: str
    :return: paths
    :rtype: dict
    """
    processor_path = get_processor_path(filename)
    return {'tetrode_data': processor_path + '/downsampled_tetrode_data/',
            'aux_data': processor_path + '/downsampled_AUX_data/',
            'timestamps': processor_path + '/downsampled_timestamps/',
            'info': processor_path + '/downsampling_info/'}


def check_if_downsampled_data_available(filename):
    """
    Checks if downsampled data is available in the NWB file.

    :param filename: path to NWB file
    :type filename: str
    :return: available
    :rtype: bool
    """
    paths = get_downsampled_data_paths(filename)
    with h5py.File(filename, 'r') as h5file:
        # START Workaround for older downsampled datasets
        if '/acquisition/timeseries/recording1/continuous/processor102_100/tetrode_lowpass' in h5file:
            return True
        # END Workaround for older downsampled datasets
        for path in [paths[key] for key in paths]:
            if not (path in h5file):
                return False
        if h5file[paths['tetrode_data']].shape[0] == 0:
            return False
        if h5file[paths['tetrode_data']].shape[0] \
                != h5file[paths['timestamps']].shape[0] \
                != h5file[paths['aux_data']].shape[0]:
            return False

    return True


def get_raw_data_paths(filename):
    """
    Returns paths to downsampled data in NWB file.

    :param filename: path to NWB file
    :type filename: str
    :return: paths
    :rtype: dict
    """
    processor_path = get_processor_path(filename)
    return {'continuous': processor_path + '/data',
            'timestamps': processor_path + '/timestamps'}


def check_if_raw_data_available(filename):
    """
    Returns paths to raw data in NWB file.

    :param filename:
    :type filename: str
    :return: paths
    :rtype: dict
    """
    paths = get_raw_data_paths(filename)
    if all([check_if_path_exists(filename, paths[key]) for key in paths]):
        return True
    else:
        return False


def save_downsampling_info_to_disk(filename, info):
    # Get paths to respective dataset locations
    paths = get_downsampled_data_paths(filename)
    # Ensure dictionary fields are in correct format
    info = {'original_sampling_rate': np.int64(info['original_sampling_rate']),
            'downsampled_sampling_rate': np.int64(info['downsampled_sampling_rate']),
            'downsampled_channels': np.array(info['downsampled_channels'], dtype=np.int64)}
    # Write data to disk
    with h5py.File(filename, 'r+') as h5file:
        recursively_save_dict_contents_to_group(h5file, paths['info'], info)


def save_downsampled_data_to_disk(filename, tetrode_data, timestamps, aux_data, info):
    # Get paths to respective dataset locations
    paths = get_downsampled_data_paths(filename)
    # Write data to disk
    save_downsampling_info_to_disk(filename, info)
    with h5py.File(filename, 'r+') as h5file:
        h5file[paths['tetrode_data']] = tetrode_data
        h5file[paths['timestamps']] = timestamps
        h5file[paths['aux_data']] = aux_data


def delete_raw_data(filename, only_if_downsampled_data_available=True):
    if only_if_downsampled_data_available:
        if not check_if_downsampled_data_available(filename):
            print('Warning', 'Downsampled data not available in NWB file. Raw data deletion aborted.')
            return None
    if not check_if_raw_data_available(filename):
        print('Warning', 'Raw data not available to be deleted in: ' + filename)
    else:
        raw_data_paths = get_raw_data_paths(filename)
        with h5py.File(filename,'r+') as h5file:
            for path in [raw_data_paths[key] for key in raw_data_paths]:
                del h5file[path]


def repack_NWB_file(filename, replace_original=True, check_validity_with_downsampled_data=True):
    # Create a repacked copy of the file
    os.system('h5repack ' + filename + ' ' + (filename + '.repacked'))
    # Check that the new file is not corrupted
    if check_validity_with_downsampled_data:
        if not check_if_downsampled_data_available(filename):
            raise Exception('Downsampled data cannot be found in repacked file. Original file not replaced.')
    # Replace original file with repacked file
    if replace_original:
        os.system('mv ' + (filename + '.repacked') + ' ' + filename)


def repack_all_nwb_files_in_directory_tree(folder_path, replace_original=True,
                                           check_validity_with_downsampled_data=True):
    # Commence directory walk
    for dir_name, subdirList, fileList in os.walk(folder_path):
        for fname in fileList:
            fpath = os.path.join(dir_name, fname)
            if fname == 'experiment_1.nwb':
                print('Repacking file {}'.format(fpath))
                repack_NWB_file(fpath, replace_original=replace_original,
                                check_validity_with_downsampled_data=check_validity_with_downsampled_data)


def list_AUX_channels(filename, n_tetrodes):
    data = load_continuous(filename)
    n_channels = data['continuous'].shape[1]
    data['file_handle'].close()
    aux_chan_list = range(n_tetrodes * 4 - 1, n_channels)

    return aux_chan_list


def load_continuous(filename):
    # Load data file
    h5file = h5py.File(filename, 'r')
    # Load timestamps and continuous data
    recordingKey = get_recordingKey(filename)
    processorKey = get_processorKey(filename)
    path = '/acquisition/timeseries/' + recordingKey + '/continuous/' + processorKey
    if check_if_path_exists(filename, path + '/data'):
        continuous = h5file[path + '/data'] # not converted to microvolts!!!! need to multiply by 0.195
        timestamps = h5file[path + '/timestamps']
        data = {'continuous': continuous, 'timestamps': timestamps, 'file_handle': h5file}
    else:
        data = None

    return data


def load_raw_data_timestamps_as_array(filename):
    data = load_continuous(filename)
    timestamps = np.array(data['timestamps']).squeeze()
    data['file_handle'].close()

    return timestamps


def load_data_columns_as_array(filename, data_path, first_column, last_column):
    """
    Loads a contiguous columns of dataset efficiently from HDF5 dataset.
    """
    with h5py.File(filename, 'r') as h5file:
        data = h5file[data_path]
        data = h5file[data_path][:, first_column:last_column]

    return data


def load_data_as_array(filename, data_path, columns):
    """
    Fast way of reading a single column or a set of columns.
    
    filename - str - full path to file
    columns  - list - column numbers to include (starting from 0).
               Single column can be given as a single list element or int.
               Columns in the list must be in sorted (ascending) order.
    """
    # Make columns variable into a list if int given
    if isinstance(columns, int):
        columns = [columns]
    # Check that all elements of columns are integers
    if isinstance(columns, list):
        for column in columns:
            if not isinstance(column, int):
                raise ValueError('columns argument must be a list of int values.')
    else:
        raise ValueError('columns argument must be list or int.')
    # Check that column number are sorted
    if sorted(columns) != columns:
        raise ValueError('columns was not in sorted (ascending) order.')
    # Check that data is available, otherwise return None
    if not check_if_path_exists(filename, data_path):
        raise ValueError('File ' + filename + '\n'
                         + 'Does not contain path ' + data_path)
    # Find contiguous column groups
    current_column = columns[0]
    column_groups = [current_column]
    for i in range(1, len(columns)):
        if (columns[i] - columns[i - 1]) == 1:
            column_groups.append(current_column)
        else:
            column_groups.append(columns[i])
            current_column = columns[i]
    # Find start and end column numbers for contiguous groups
    column_ranges = []
    for first_channel in sorted(set(column_groups)):
        last_channel = first_channel + column_groups.count(first_channel)
        column_ranges.append((first_channel, last_channel))
    # Get contiguous column segments for each group
    column_group_data = []
    for column_range in column_ranges:
        column_group_data.append(
            load_data_columns_as_array(filename, data_path, *column_range))
    # Concatenate column groups
    data = np.concatenate(column_group_data, axis=1)

    return data


def load_continuous_as_array(filename, channels):
    """
    Fast way of reading a single channel or a set of channels.
    
    filename - str - full path to file
    channels - list - channel numbers to include (starting from 0).
               Single channel can be given as a single list element or int.
               Channels in the list must be in sorted (ascending) order.
    """
    # Generate path to raw continuous data
    root_path = '/acquisition/timeseries/' + get_recordingKey(filename) \
                + '/continuous/' + get_processorKey(filename)
    data_path = root_path + '/data'
    timestamps_path = root_path + '/timestamps'
    # Check that data is available, otherwise return None
    if not check_if_path_exists(filename, data_path):
        return None
    if not check_if_path_exists(filename, timestamps_path):
        return None
    # Load continuous data
    continuous = load_data_as_array(filename, data_path, channels)
    # Load timestamps for data
    with h5py.File(filename, 'r') as h5file:
        timestamps = np.array(h5file[timestamps_path])
    # Arrange output into a dictionary
    data = {'continuous': continuous, 'timestamps': timestamps}

    return data


def remove_surrounding_binary_markers(text):
    if text.startswith("b'"):
        text = text[2:]
    if text.endswith("'"):
        text = text[:-1]
    return text


def get_downsampling_info_old(filename):
    # Generate path to downsampling data info
    root_path = '/acquisition/timeseries/' + get_recordingKey(filename) \
                + '/continuous/' + get_processorKey(filename)
    data_path = root_path + '/downsampling_info'
    # Load info from file
    with h5py.File(filename, 'r') as h5file:
        data = h5file[data_path]
        data = [str(i) for i in data]
    # Remove b'x' markers from strings if present. Python 3 change.
    data = list(map(remove_surrounding_binary_markers, data))
    # Parse elements in loaded data
    info_dict = {}
    for x in data:
        key, value = x.split(' ')
        if key == 'original_sampling_rate':
            info_dict[key] = np.int64(value)
        elif key == 'downsampled_sampling_rate':
            info_dict[key] = np.int64(value)
        elif key == 'downsampled_channels':
            info_dict[key] = np.array(list(map(int, value.split(',')))).astype(np.int64)

    return info_dict


def get_downsampling_info(filename):
    root_path = '/acquisition/timeseries/' + get_recordingKey(filename) \
                + '/continuous/' + get_processorKey(filename)
    data_path = root_path + '/downsampling_info/'
    with h5py.File(filename, 'r') as h5file:
        return recursively_load_dict_contents_from_group(h5file, data_path)


def load_downsampled_tetrode_data_as_array(filename, tetrode_nrs):
    """
    Returns a dict with downsampled continuous data for requested tetrodes
    
    filename                    - str - full path to file
    tetrode_nrs                 - list - tetrode numbers to include (starting from 0).
                                Single tetrode can be given as a single list element or int.
                                Tetrode numbers in the list must be in sorted (ascending) order.
                                If data is not available for a given tetrode number, error is raised.
    """
    # Generate path to raw continuous data
    root_path = '/acquisition/timeseries/' + get_recordingKey(filename) \
                + '/continuous/' + get_processorKey(filename)
    data_path = root_path + '/downsampled_tetrode_data'
    timestamps_path = root_path + '/downsampled_timestamps'
    # Check that data is available, otherwise return None
    if not check_if_path_exists(filename, data_path):
        return None
    if not check_if_path_exists(filename, timestamps_path):
        return None
    # Get info on downsampled data
    info = get_downsampling_info(filename)
    sampling_rate = int(info['downsampled_sampling_rate'])
    downsampled_channels = list(info['downsampled_channels'])
    # Map tetrode_nrs elements to columns in downsampled_tetrode_data
    columns = []
    channels_used = []
    tetrode_nrs_remaining = copy(tetrode_nrs)
    for tetrode_nr in tetrode_nrs:
        for chan in tetrode_channels(tetrode_nr):
            if chan in downsampled_channels:
                columns.append(downsampled_channels.index(chan))
                channels_used.append(chan)
                tetrode_nrs_remaining.pop(tetrode_nrs_remaining.index(tetrode_nr))
                break
    # Check that all tetrode numbers were mapped
    if len(tetrode_nrs_remaining) > 0:
        raise Exception('The following tetrodes were not represented in downsampled data\n' \
                        + ','.join(list(map(str, tetrode_nrs_remaining))))
    # Load continuous data
    continuous = load_data_as_array(filename, data_path, columns)
    # Load timestamps for data
    with h5py.File(filename, 'r') as h5file:
        timestamps = np.array(h5file[timestamps_path])
    # Arrange output into a dictionary
    data = {'continuous': continuous, 'timestamps': timestamps,
            'tetrode_nrs': tetrode_nrs, 'channels': channels_used,
            'sampling_rate': sampling_rate}

    return data


def empty_spike_data():
    """
    Creates a fake waveforms of 0 values and at timepoint 0
    """
    waveforms = np.zeros((1,4,40), dtype=np.int16)
    timestamps = np.array([0], dtype=np.float64)

    return {'waveforms': waveforms, 'timestamps': timestamps}


def get_tetrode_nrs_if_spikes_available(filename, spike_name='spikes'):
    """
    Returns a list of tetrode numbers if spikes available in NWB file.
    """
    spikes_path = '/acquisition/timeseries/' + get_recordingKey(filename) + '/' + spike_name + '/'
    # Get tetrode keys if available
    with h5py.File(filename, 'r') as h5file:
        if not (spikes_path in h5file):
            # Return empty list if spikes data not available
            return []
        tetrode_keys = list(h5file[spikes_path].keys())
    # Return empty list if spikes not available on any tetrode
    if len(tetrode_keys) == 0:
        return []
    # Extract tetrode numbers
    tetrode_nrs = []
    for tetrode_key in tetrode_keys:
        tetrode_nrs.append(int(tetrode_key[9:]) - 1)
    # Sort tetrode numbers in ascending order
    tetrode_nrs.sort()

    return tetrode_nrs


def construct_paths_to_tetrode_spike_data(filename, tetrode_nrs, spike_name='spikes'):
    spikes_path = '/acquisition/timeseries/' + get_recordingKey(filename) + '/' + spike_name + '/'
    return [(spikes_path + 'electrode' + str(tetrode_nr + 1) + '/') for tetrode_nr in tetrode_nrs]


def count_spikes(filename, tetrode_nrs, spike_name='spikes', use_idx_keep=False):
    """
    :param filename: full path to NWB file
    :type filename: str
    :param tetrode_nrs: tetrode numbers to count spikes for
    :type tetrode_nrs: list
    :param spike_name: type of spikes to look for (field in NWB file)
    :type spike_name: str
    :param use_idx_keep: If False (default) all spikes are counted, otherwise only filtered spikes are counted
    :type use_idx_keep: bool
    :return: total number of spikes on each tetrode
    :rtype: list
    """
    tetrode_paths = construct_paths_to_tetrode_spike_data(filename, tetrode_nrs, spike_name=spike_name)
    count = []
    with h5py.File(filename, 'r') as h5file:
        for tetrode_path in tetrode_paths:
            if use_idx_keep:
                count.append(sum(np.array(h5file[tetrode_path + 'idx_keep'][()]).squeeze()))
            else:
                count.append(h5file[tetrode_path + 'timestamps/'].shape[0])

    return count


def load_spikes(filename, spike_name='spikes', tetrode_nrs=None, use_idx_keep=False,
                use_badChan=False, no_waveforms=False, clustering_name=None, verbose=True):
    """
    Inputs:
        filename - pointer to NWB file to load
        tetrode_nrs [list]    - can be a list of tetrodes to load (from 0)
        use_idx_keep [bool]   - if True, only outputs spikes according to idx_keep of tetrode, if available
        use_badChan [bool]    - if True, sets all spikes on badChannels to 0
        no_waveforms [bool]   - if True, waveforms are not loaded
        clustering_name [str] - if specified, clusterID will be loaded from:
                              electrode[nr]/clustering/clustering_name
        verbose [bool]        - prints out loading progress bar if True (default)
    Output:
        List of dictionaries for each tetrode in correct order where:
        List is empty, if no spike data detected
        'waveforms' is a list of tetrode waveforms in the order of channels
        'timestamps' is a list of spike detection timestamps corresponding to 'waveforms'
        If available, two more variables will be in the dictionary
        'idx_keep' is boolan index for 'waveforms' and 'timestamps' indicating the spikes
            that are to be used for further processing (based on filtering for artifacts etc)
        'clusterIDs' is the cluster identities of spikes in 'waveforms'['idx_keep',:,:]
    """
    # If not provided, get tetrode_nrs
    if tetrode_nrs is None:
        tetrode_nrs = get_tetrode_nrs_if_spikes_available(filename, spike_name=spike_name)
    tetrode_paths = construct_paths_to_tetrode_spike_data(filename, tetrode_nrs, spike_name=spike_name)
    with h5py.File(filename, 'r') as h5file:
        # Put waveforms and timestamps into a list of dictionaries in correct order
        data = []
        if verbose:
            print('Loading tetrodes from {}'.format(filename))
        iterable = zip(tetrode_nrs, tetrode_paths)
        for nr_tetrode, tetrode_path in (tqdm(iterable, total=len(tetrode_nrs)) if verbose else iterable):
            # Load waveforms and timestamps
            if no_waveforms:
                waveforms = empty_spike_data()['waveforms']
            else:
                waveforms = h5file[tetrode_path + 'data/'][()]
            timestamps = h5file[tetrode_path + 'timestamps/'][()]
            if not isinstance(timestamps, np.ndarray):
                timestamps = np.array([timestamps])
            if waveforms.shape[0] == 0:
                # If no waveforms are available, enter one waveform of zeros at timepoint zero
                waveforms = empty_spike_data()['waveforms']
                timestamps = empty_spike_data()['timestamps']
            # Arrange waveforms, timestamps and nr_tetrode into a dictionary
            tet_data = {'waveforms': waveforms,
                        'timestamps': timestamps,
                        'nr_tetrode': nr_tetrode}
            # Include idx_keep if available
            idx_keep_path = tetrode_path + 'idx_keep'
            if idx_keep_path in h5file:
                tet_data['idx_keep'] = np.array(h5file[idx_keep_path][()])
                if use_idx_keep:
                    # If requested, filter wavefoms and timestamps based on idx_keep
                    if np.sum(tet_data['idx_keep']) == 0:
                        tet_data['waveforms'] = empty_spike_data()['waveforms']
                        tet_data['timestamps'] = empty_spike_data()['timestamps']
                    else:
                        if not no_waveforms:
                            tet_data['waveforms'] = tet_data['waveforms'][tet_data['idx_keep'], :, :]
                        tet_data['timestamps'] = tet_data['timestamps'][tet_data['idx_keep']]
            # Include clusterIDs if available
            if clustering_name is None:
                clusterIDs_path = tetrode_path + 'clusterIDs'
            else:
                clusterIDs_path = tetrode_path + '/clustering/' + clustering_name
            if clusterIDs_path in h5file:
                tet_data['clusterIDs'] = np.int16(h5file[clusterIDs_path][()]).squeeze()
            # Set spikes to zeros for channels in badChan list if requested
            if use_badChan and not no_waveforms:
                badChan = listBadChannels(filename)
                if len(badChan) > 0:
                    for nchan in tetrode_channels(nr_tetrode):
                        if nchan in badChan:
                            tet_data['waveforms'][:, np.mod(nchan, 4), :] = 0
            data.append(tet_data)
        
        return data

def save_spikes(filename, tetrode_nr, data, timestamps, spike_name='spikes', overwrite=False):
    """
    Stores spike data in NWB file in the same format as with OpenEphysGUI.
    tetrode_nr=0 for first tetrode.
    """
    if data.dtype != np.int16:
        raise ValueError('Waveforms are not int16.')
    if timestamps.dtype != np.float64:
        raise ValueError('Timestamps are not float64.')
    recordingKey = get_recordingKey(filename)
    path = '/acquisition/timeseries/' + get_recordingKey(filename) + '/' + spike_name + '/' + \
           'electrode' + str(tetrode_nr + 1) + '/'
    if check_if_path_exists(filename, path):
        if overwrite:
            # If overwrite is true, path is first cleared
            with h5py.File(filename, 'r+') as h5file:
                del h5file[path]
        else:
            raise Exception('Spikes already in file and overwrite not requested.\n' \
                            + 'File: ' + filename + '\n' \
                            + 'path: ' + path)
    with h5py.File(filename, 'r+') as h5file:
        h5file[path + 'data'] = data
        h5file[path + 'timestamps'] = np.float64(timestamps).squeeze()

def processing_method_and_spike_name_combinations():
    """
    Outputs a list of potential processing_method and spike_name combinations
    """
    processing_methods = ['klustakwik', 'klustakwik_raw', 'kilosort']
    spike_names = ['spikes', 'spikes_raw', 'spikes_kilosort']

    return processing_methods, spike_names

def get_spike_name_for_processing_method(processing_method):
    processing_methods, spike_names = processing_method_and_spike_name_combinations()
    spike_name = spike_names[processing_methods.index(processing_method)]

    return spike_name

def load_events(filename):
    # Outputs a dictionary timestamps and eventIDs for TTL signals received
    # timestamps are in seconds, aligned to timestamps of continuous recording
    # eventIDs indicate TTL channel number (starting from 1) and are positive for rising signals

    # Load data file
    recordingKey = get_recordingKey(filename)
    with h5py.File(filename, 'r') as h5file:
        # Load timestamps and TLL signal info
        timestamps = h5file['acquisition']['timeseries'][recordingKey]['events']['ttl1']['timestamps'][()]
        eventID = h5file['acquisition']['timeseries'][recordingKey]['events']['ttl1']['data'][()]
        data = {'eventID': eventID, 'timestamps': timestamps}

    return data

def load_GlobalClock_timestamps(filename, GlobalClock_TTL_channel=1):
    """
    Returns timestamps of GlobalClock TTL pulses.
    """
    data = load_events(filename)
    return data['timestamps'][data['eventID'] == GlobalClock_TTL_channel]


def load_network_events(filename):
    """returns network_events_data

    Extracts the list of network messages from NWB file 
    and returns it along with corresponding timestamps
    in dictionary with keys ['messages', 'timestamps']
    
    'messages' - list of str
    
    'timestamps' - list of float

    :param filename: full path to NWB file
    :type filename: str
    :return: network_events_data
    :rtype: dict
    """
    # Load data file
    recordingKey = get_recordingKey(filename)
    with h5py.File(filename, 'r') as h5file:
        # Load timestamps and messages
        timestamps = h5file['acquisition']['timeseries'][recordingKey]['events']['text1']['timestamps'][()]
        messages = h5file['acquisition']['timeseries'][recordingKey]['events']['text1']['data'][()]
    messages = [x.decode('utf-8') for x in messages]
    timestamps = [float(x) for x in timestamps]

    data = {'messages': messages, 'timestamps': timestamps}

    return data


def check_if_path_exists(filename, path):
    with h5py.File(filename, 'r') as h5file:
        return path in h5file


def save_list_of_dicts_to_group(h5file, path, dlist, overwrite=False, list_suffix='_NWBLIST'):
    # Check that all elements are dictionaries
    for dic in dlist:
        if not isinstance(dic, dict):
            raise Exception('List elements must be dictionaries')
    # Write elements to file
    for i, dic in enumerate(dlist):
        recursively_save_dict_contents_to_group(h5file, (path + str(i) + '/'), dic,
                                                overwrite=overwrite, list_suffix=list_suffix)


def recursively_save_dict_contents_to_group(h5file, path, dic, overwrite=False, list_suffix='_NWBLIST', verbose=False):
    """
    h5file - h5py.File
    path   - str       - path to group in h5file. Must end with '/'
    overwrite - bool   - any dictionary elements or lists that already exist are overwritten.
                         Default is False, if elements already exist in NWB file, error is raised.
    list_suffix - str  - suffix used to highlight paths created from lists of dictionaries.
                         Must be consistent when saving and loading data.
    verbose - bool     - If True (default is False), h5file path used is printed for each recursion

    Only works with: numpy arrays, numpy int64 or float64, strings, bytes, lists of strings and dictionaries these are contained in.
    Also works with lists dictionaries as part of the hierachy.
    Long lists of dictionaries are discouraged, as individual groups are created for each element.
    """
    if verbose:
        print(path)
    if len(dic) == 0:
        if path in h5file:
            del h5file[path]
        h5file.create_group(path)
    for key, item in dic.items():
        if isinstance(item, (int, float)):
            item = np.array(item)
        if isinstance(item, (np.ndarray, np.int64, np.float64, str, bytes)):
            if overwrite:
                if path + key in h5file:
                    del h5file[path + key]
            h5file[path + key] = item
        elif isinstance(item, dict):
            recursively_save_dict_contents_to_group(h5file, path + key + '/', item,
                                                    overwrite=overwrite, list_suffix=list_suffix,
                                                    verbose=verbose)
        elif isinstance(item, list):
            if all(isinstance(i, str) for i in item):
                if overwrite:
                    if path + key in h5file:
                        del h5file[path + key]
                asciiList = [n.encode("ascii", "ignore") for n in item]
                h5file[path + key] = h5file.create_dataset(None, (len(asciiList),),'S100', asciiList)
            else:
                if overwrite:
                    if path + key + list_suffix in h5file:
                        del h5file[path + key + list_suffix]
                save_list_of_dicts_to_group(h5file, path + key + list_suffix + '/', item, 
                                            overwrite=overwrite, list_suffix=list_suffix)
        elif item is None:
            h5file.create_group(path + key)
        else:
            raise ValueError('Cannot save %s type'%type(item) + ' from ' + path + key)


def convert_bytes_to_string(b):
    """
    If input is bytes, returns str decoded with utf-8

    :param b:
    :type b: bytes
    :return: string decoded with utf-8 if input is bytes object, otherwise returns unchanged input
    :rtype: str
    """
    if isinstance(b, bytes):
        if sys.version_info >= (3, 0):
            return str(b, 'utf-8')
        else:
            return str(b.decode('utf-8'))
    else:
        return b


def load_list_of_dicts_from_group(h5file, path, list_suffix='_NWBLIST', ignore=()):
    # Load all elements on this path
    items = []
    for key in list(h5file[path].keys()):
        items.append(
            (int(key), recursively_load_dict_contents_from_group(h5file, path + key + '/', 
                                                                 list_suffix=list_suffix,
                                                                 ignore=ignore))
        )
    # Create a list from items sorted by group keys
    ans = [item for _, item in sorted(items)]

    return ans


def recursively_load_dict_contents_from_group(h5file, path, list_suffix='_NWBLIST', ignore=()):
    """
    Returns value at path if it has no further items

    h5file - h5py.File
    path   - str       - path to group in h5file. Must end with '/'
    list_suffix - str  - suffix used to highlight paths created from lists of dictionaries.
                         Must be consistent when saving and loading data.
    ignore - tuple     - paths including elements matching any element in this tuple return None
    """
    if not path.endswith('/'):
        raise ValueError('Input path must end with "/"')

    if path.split('/')[-2] in ignore or path.split('/')[-2][:-len(list_suffix)] in ignore:
        ans = None

    elif path[:-1].endswith(list_suffix):
        ans = load_list_of_dicts_from_group(h5file, path, list_suffix=list_suffix,
                                            ignore=ignore)
    elif hasattr(h5file[path], 'items'):

        ans = {}
        for key, item in h5file[path].items():

            if key.endswith(list_suffix):
                ans[str(key)[:-len(list_suffix)]] = load_list_of_dicts_from_group(
                    h5file, path + key + '/', list_suffix=list_suffix,
                    ignore=ignore
                )

            elif isinstance(item, h5py._hl.dataset.Dataset):
                if 'S100' == item.dtype:
                    tmp = list(item[()])
                    ans[str(key)] = [convert_bytes_to_string(i) for i in tmp]
                elif item.dtype == 'bool' and item.ndim == 0:
                    ans[str(key)] = np.array(bool(item[()]))
                else:
                    ans[str(key)] = convert_bytes_to_string(item[()])

            elif isinstance(item, h5py._hl.group.Group):
                ans[str(key)] = recursively_load_dict_contents_from_group(h5file, path + key + '/',
                                                                          ignore=ignore)

    else:
        ans = convert_bytes_to_string(h5file[path][()])

    return ans


def save_settings(filename, Settings, path='/'):
    """
    Writes into an existing file if path is not yet used.
    Creates a new file if filename does not exist.
    Only works with: numpy arrays, numpy int64 or float64, strings, bytes, lists of strings and dictionaries these are contained in.
    To save specific subsetting, e.g. TaskSettings, use:
        Settings=TaskSetttings, path='/TaskSettings/'
    """
    full_path = '/general/data_collection/Settings' + path
    if os.path.isfile(filename):
        write_method = 'r+'
    else:
        write_method = 'w'
    with h5py.File(filename, write_method) as h5file:
        recursively_save_dict_contents_to_group(h5file, full_path, Settings)

def load_settings(filename, path='/', ignore=()):
    """
    By default loads all settings from path
        '/general/data_collection/Settings/'
    or for example to load animal ID, use:
        path='/General/animal/'

    ignore - tuple - any paths including any element of ignore are returned as None
    """
    full_path = '/general/data_collection/Settings' + path
    with h5py.File(filename, 'r') as h5file:
        data = recursively_load_dict_contents_from_group(h5file, full_path, ignore=ignore)

    return data


def check_if_settings_available(filename, path='/'):
    """
    Returns whether settings information exists in NWB file
    Specify path='/General/badChan/' to check for specific settings
    """
    full_path = '/general/data_collection/Settings' + path
    with h5py.File(filename, 'r') as h5file:
        return full_path in h5file


def save_analysis(filename, data, overwrite=False, complete_overwrite=False, verbose=False):
    """Stores analysis results from nested dictionary to /analysis path in NWB file.

    See :py:func:`NWBio.recursively_save_dict_contents_to_group` for details on supported data structures.

    :param str filename: path to NWB file
    :param dict data: analysis data to be stored in NWB file
    :param bool overwrite: if True, any existing data at same dictionary keys
                           as in previously saved data is overwritten.
                           Default is False.
    :param bool complete_overwrite: if True, all previous analysis data is discarded before writing.
                                    Default is False.
    :param bool verbose: if True (default is False), the path in file for each element is printed.
    """
    with h5py.File(filename, 'r+') as h5file:
        if complete_overwrite:
            del h5file['/analysis']
        recursively_save_dict_contents_to_group(h5file, '/analysis/', data, overwrite=overwrite, verbose=verbose)


def load_analysis(filename, ignore=()):
    """Loads analysis results from /analysis path in NWB file into a dictionary.

    :param str filename: path to NWB file
    :param tuple ignore: paths containing any element of ignore are terminated with None.
        In the output dictionary any elements downstream of a key matching any element of ignore
        is not loaded and dictionary tree is terminated at that point with value None.
    """
    with h5py.File(filename, 'r') as h5file:
        return recursively_load_dict_contents_from_group(h5file, '/analysis/', ignore=ignore)


def listBadChannels(filename):
    if check_if_settings_available(filename,'/General/badChan/'):
        badChanString = load_settings(filename,'/General/badChan/')
        # Separate input string into a list using ',' as deliminaters
        if badChanString.find(',') > -1: # If more than one channel specified
            # Find all values tetrode and channel values listed
            badChanStringList = badChanString.split(',')
        else:
            badChanStringList = [badChanString]
        # Identify any ranges specified with '-' and append these channels to the list
        for chanString in badChanStringList:
            if chanString.find('-') > -1:
                chan_from = chanString[:chanString.find('-')]
                chan_to = chanString[chanString.find('-') + 1:]
                for nchan in range(int(chan_to) - int(chan_from) + 1):
                    badChanStringList.append(str(nchan + int(chan_from)))
                badChanStringList.remove(chanString) # Remove the '-' containing list element
        # Reorder list of bad channels
        badChanStringList.sort(key=int)
        badChan = list(np.array(list(map(int, badChanStringList))) - 1)
    else:
        badChan = []

    return badChan

def save_tracking_data(filename, TrackingData, ProcessedPos=False, overwrite=False):
    """
    TrackingData is expected as dictionary with keys for each source ID
    If saving processed data, TrackingData is expected to be numpy array
        Use ProcessedPos=True to store processed data
        Use overwrite=True to force overwriting existing processed data
    """
    if os.path.isfile(filename):
        write_method = 'r+'
    else:
        write_method = 'w'
    recordingKey = get_recordingKey(filename)
    with h5py.File(filename, write_method) as h5file:
        full_path = '/acquisition/timeseries/' + recordingKey + '/tracking/'
        if not ProcessedPos:
            recursively_save_dict_contents_to_group(h5file, full_path, TrackingData)
        elif ProcessedPos:
            processed_pos_path = full_path + 'ProcessedPos/'
            # If overwrite is true, path is first cleared
            if overwrite:
                if full_path in h5file and 'ProcessedPos' in list(h5file[full_path].keys()):
                    del h5file[processed_pos_path]
            h5file[processed_pos_path] = TrackingData


def get_recording_cameraIDs(filename):
    path = '/general/data_collection/Settings/CameraSettings/CameraSpecific'
    with h5py.File(filename, 'r') as h5file:
        if path in h5file:
            return list(h5file[path].keys())


def load_raw_tracking_data(filename, cameraID, specific_path=None):
    path = '/acquisition/timeseries/' + get_recordingKey(filename) + '/tracking/' + cameraID + '/'
    if not (specific_path is None):
        path = path + '/' + specific_path + '/'
    with h5py.File(filename, 'r') as h5file:
        if path in h5file:
            return recursively_load_dict_contents_from_group(h5file, path)

def load_processed_tracking_data(filename, subset='ProcessedPos'):
    path = '/acquisition/timeseries/' + get_recordingKey(filename) + '/tracking/'
    path = path + subset
    with h5py.File(filename, 'r') as h5file:
        return np.array(h5file[path][()])

def get_processed_tracking_data_timestamp_edges(filename, subset='ProcessedPos'):
    if check_if_processed_position_data_available(filename):
        data = load_processed_tracking_data(filename, subset=subset)
        edges = [data[0, 0], data[-1, 0]]
    else:
        print('Warning! ProcessedPos not available. Using continuous data timestamps')
        h5file = h5py.File(filename, 'r')
        recordingKey = get_recordingKey(filename)
        processorKey = get_processorKey(filename)
        path = '/acquisition/timeseries/' + recordingKey + '/continuous/' + processorKey + '/timestamps'
        edges = [h5file[path][0], h5file[path][-1]]
        h5file.close()

    return edges

def check_if_tracking_data_available(filename):
    if check_if_settings_available(filename, path='/General/Tracking/'):
        return load_settings(filename, path='/General/Tracking/')
    else:
        return False

def check_if_processed_position_data_available(filename, subset='ProcessedPos'):
    path = '/acquisition/timeseries/' + get_recordingKey(filename) + '/tracking/'
    path = path + subset
    return check_if_path_exists(filename, path)

def check_if_binary_pos(filename):
    # Checks if binary position data exists in NWB file
    path = '/acquisition/timeseries/' + get_recordingKey(filename) + '/events/binary1/'
    return check_if_path_exists(filename, path)


def save_tetrode_idx_keep(filename, ntet, idx_keep, spike_name='spikes', overwrite=False):
    path = '/acquisition/timeseries/' + get_recordingKey(filename) + '/' + spike_name + '/' + \
           'electrode' + str(ntet + 1) + '/idx_keep'
    with h5py.File(filename, 'r+') as h5file:
        if path in h5file:
            if overwrite:
                del h5file[path]
            else:
                raise ValueError('Tetrode ' + str(ntet + 1) + ' idx_keep already exists in ' + filename)
        h5file[path] = idx_keep

def save_tetrode_clusterIDs(filename, ntet, clusterIDs, spike_name='spikes', overwrite=False):
    path = '/acquisition/timeseries/' + get_recordingKey(filename) + '/' + spike_name + '/' + \
           'electrode' + str(ntet + 1) + '/clusterIDs'
    with h5py.File(filename, 'r+') as h5file:
        if path in h5file:
            if overwrite:
                del h5file[path]
            else:
                raise ValueError('Tetrode ' + str(ntet + 1) + ' clusterIDs already exists in ' + filename)
        h5file[path] = np.int16(clusterIDs).squeeze()

def fill_empty_dictionary_from_source(selection, src_dict):
    """
    Populates a dictionary with None values with values from a source
    dictionary with identical structure.
    """
    dst_dict = copy(selection)
    for key, item in dst_dict.items():
        if isinstance(item, dict):
            dst_dict[key] = fill_empty_dictionary_from_source(item, src_dict[key])
        elif item is None:
            dst_dict[key] = src_dict[key]
        else:
            raise ValueError('Destination dictionary has incorrect.')

    return dst_dict


def get_recording_start_timestamp_offset(filename):
    """Returns the first timestamp of raw or downsampled continuous data.

    :param str filename: path to NWB file
    :return: first timestamp of continuous data
    :rtype: float
    """
    if check_if_raw_data_available(filename):
        path = get_raw_data_paths(filename)['timestamps']
    elif check_if_downsampled_data_available(filename):
        path = get_downsampled_data_paths(filename)['timestamps']
    else:
        raise Exception('NWB file does not contain raw or downsampled data ' + filename)
    with h5py.File(filename, 'r') as h5file:
        return float(h5file[path][0:1])


def get_recording_full_duration(filename):
    """Returns the total duration from first to last timestamp of
    raw or downsampled continuous data.

    :param str filename: path to NWB file
    :return: total duration from first to last timestamp of continuous data
    :rtype: float
    """
    if check_if_raw_data_available(filename):
        path = get_raw_data_paths(filename)['timestamps']
    elif check_if_downsampled_data_available(filename):
        path = get_downsampled_data_paths(filename)['timestamps']
    else:
        raise Exception('NWB file does not contain raw or downsampled data ' + filename)
    with h5py.File(filename, 'r') as h5file:
        return float(h5file[path][-1]) - float(h5file[path][0:1])


def import_task_specific_log_parser(task_name):
    """
    Returns LogParser module for the specific task.

    :param task_name: name of the task
    :type task_name: str
    :return: TaskLogParser
    :rtype: module
    """
    if task_name == 'Pellets_and_Rep_Milk_Task':  # Temporary workaround to function with older files
        task_name = 'Pellets_and_Rep_Milk'
    try:
        return importlib.import_module('.Tasks.' + task_name + '.LogParser', package='openEPhys_DACQ')
    except ModuleNotFoundError:
        print('Task {} LogParser not found. Returning None.'.format(task_name))
        return None


def load_task_name(filename):
    """
    Returns the name of the task active in the recording.

    :param filename: absolute path to NWB recording file
    :type filename: str
    :return: task_name
    :rtype: str
    """
    return load_settings(filename, path='/TaskSettings/name/')


def get_recording_log_parser(filename, final_timestamp=None):
    """Finds task specific LogParser class and returns it initialized
    with network events from that recording.

    :param str filename:
    :return: Task specific log parser initialized with network events
    :rtype: LogParser class
    """
    task_log_parser = import_task_specific_log_parser(load_task_name(filename))
    if task_log_parser is None:
        return None
    else:
        return task_log_parser.LogParser(task_settings=load_settings(filename, path='/TaskSettings/'),
                                         final_timestamp=final_timestamp,
                                         **load_network_events(filename))


def get_channel_map(filename):
    return load_settings(filename, '/General/channel_map/')


def list_tetrode_nrs_for_area_channel_map(area_channel_map):
    return list(set([channels_tetrode(chan) for chan in list(area_channel_map['list'])]))


def get_channel_map_with_tetrode_nrs(filename):
    channel_map = get_channel_map(filename)
    for area in channel_map:
        channel_map[area]['tetrode_nrs'] = list_tetrode_nrs_for_area_channel_map(channel_map[area])

    return channel_map


def check_if_channel_maps_are_same(channel_map_1, channel_map_2):
    """
    Determines if two channel maps are identical
    """
    # Check that there are same number of areas in the dictionary
    if len(channel_map_1) != len(channel_map_2):
        return False
    # Sort the area names because dictionary is not ordered
    channel_map_1_keys = sorted(list(channel_map_1.keys()))
    channel_map_2_keys = sorted(list(channel_map_2.keys()))
    # Check that the areas have the same name
    for n_area in range(len(channel_map_1_keys)):
        if channel_map_1_keys[n_area] != channel_map_2_keys[n_area]:
            return False
    # Check that the channel lists are the same
    for area in channel_map_1_keys:
        if not all(channel_map_1[area]['list'] == channel_map_2[area]['list']):
            return False

    return True


def estimate_open_ephys_timestamps_from_other_timestamps(open_ephys_global_clock_times, other_global_clock_times,
                                                         other_times, other_times_divider=None):
    """Returns Open Ephys timestamps for each timestamp from another device by synchronising with global clock.

    Note, other times must be in same units as open_ephys_global_clock_times. Most likely seconds.
    For example, Raspberry Pi camera timestamps would need to be divided by 10 ** 6

    :param numpy.ndarray open_ephys_global_clock_times: shape (N,)
    :param numpy.ndarray other_global_clock_times: shape (M,)
    :param numpy.ndarray other_times: shape (K,)
    :param int other_times_divider: if provided, timestamps from the other devices are divided by this value
        before matching to Open Ephys time. This allows inputting timestamps from other device in original units.
        In case of Raspberry Pi camera timestamps, this value should be 10 ** 6.
        If this value is not provided, all provided timestamps must be in same units.
    :return: open_ephys_times
    :rtype: numpy.ndarray
    """

    # Crop data if more timestamps recorded on either system.
    if open_ephys_global_clock_times.size > other_global_clock_times.size:
        open_ephys_global_clock_times = open_ephys_global_clock_times[:other_global_clock_times.size]
        print('[ Warning ] OpenEphys recorded more GlobalClock TTL pulses than other system.\n' +
              'Dumping extra OpenEphys timestamps from the end.')
    elif open_ephys_global_clock_times.size < other_global_clock_times.size:
        other_global_clock_times = other_global_clock_times[:open_ephys_global_clock_times.size]
        print('[ Warning ] Other system recorded more GlobalClock TTL pulses than Open Ephys.\n' +
              'Dumping extra other system timestamps from the end.')

    # Find closest other_global_clock_times indices to each other_times
    other_times_gc_indices = closest_argmin(other_times, other_global_clock_times)

    # Compute difference from the other_global_clock_times for each value in other_times
    other_times_nearest_global_clock_times = other_global_clock_times[other_times_gc_indices]
    other_times_global_clock_delta = other_times - other_times_nearest_global_clock_times

    # Convert difference values to Open Ephys timestamp units
    if not (other_times_divider is None):
        other_times_global_clock_delta = other_times_global_clock_delta / float(other_times_divider)

    # Use other_times_global_clock_delta to estimate timestamps in OpenEphys time
    other_times_nearest_open_ephys_global_clock_times = open_ephys_global_clock_times[other_times_gc_indices]
    open_ephys_times = other_times_nearest_open_ephys_global_clock_times + other_times_global_clock_delta

    return open_ephys_times


def extract_recording_info(filename, selection='default'):
    """
    Returns recording info for the recording file.

    selection - allows specifying which data return
        'default' - some hard-coded selection of data
        'all' - all of the recording settings
        dict - a dictionary with the same exact keys and structure
               as the recording settings, with None for item values
               and missing keys for unwanted elements. The dictionary
               will be returned with None values populated by values
               from recording settings.
    """
    recording_info = {}
    if isinstance(selection, str) and selection == 'default':
        recording_info.update(load_settings(filename, '/General/'))
        del recording_info['experimenter']
        del recording_info['rec_file_path']
        del recording_info['root_folder']
        if recording_info['TaskActive']:
            recording_info.update({'TaskName': load_settings(filename, '/TaskSettings/name/')})
        for key in list(recording_info['channel_map'].keys()):
            del recording_info['channel_map'][key]['list']
        pos_edges = get_processed_tracking_data_timestamp_edges(filename)
        recording_info['duration'] = pos_edges[1] - pos_edges[0]
        recording_info['duration (min)'] = int(round((pos_edges[1] - pos_edges[0]) / 60))
        recording_info['time'] = load_settings(filename, '/Time/')
    elif isinstance(selection, str) and selection == 'all':
        recording_info = load_settings(filename)
    elif isinstance(selection, dict):
        full_recording_info = load_settings(filename)
        recording_info = fill_empty_dictionary_from_source(selection, full_recording_info)

    return recording_info


def display_recording_data(root_path, selection='default'):
    """
    Prints recording info for the whole directory tree.
    """
    for dirName, subdirList, fileList in os.walk(root_path):
        for fname in fileList:
            if fname == 'experiment_1.nwb':
                filename = os.path.join(dirName, fname)
                recording_info = extract_recording_info(filename, selection)
                print('Data on path: ' + dirName)
                pprint(recording_info)


if __name__ == '__main__':
    # Input argument handling and help info
    parser = argparse.ArgumentParser(description='Extract info from Open Ephys.')
    parser.add_argument('root_path', type=str, nargs=1, 
                        help='Root directory for recording(s)')
    args = parser.parse_args()
    # Get paths to recording files
    display_recording_data(args.root_path[0])



### These functions allow interfacing with the Raspberry Pis.

### By Sander Tanni, January 2018, UCL

import zmq
import multiprocessing
from time import time, sleep
from openEPhys_DACQ.sshScripts import ssh
import json
from threading import Lock, Thread
import numpy as np
from scipy.spatial.distance import euclidean
from openEPhys_DACQ.TrackingDataProcessing import combineCamerasData
from openEPhys_DACQ.ZMQcomms import paired_messenger, remote_object_controller
from multiprocessing.dummy import Pool as ThreadPool
from tempfile import mkdtemp
from shutil import rmtree
import os
from openEPhys_DACQ.HelperFunctions import test_pinging_address, time_string
from openEPhys_DACQ.package_configuration import package_path


def write_files_to_RPi(files, address, username='pi', verbose=False):
    """
    files - list or tuple of files to copy to home directory on the RPi
    address - RPi address
    username - RPi username to use for access
    """
    if verbose:
        qflag = ''
    else:
        qflag = '-q '
    RPi_network_path = username + '@' + address
    for file in files:
        callstr = 'scp ' + qflag + file + ' ' + RPi_network_path + ':/home/pi'
        _ = os.system(callstr)

def read_file_from_RPi(source_filename, target_filename, address, username='pi', verbose=False):
    if verbose:
        qflag = ''
    else:
        qflag = '-q '
    RPi_network_path = username + '@' + address
    src = RPi_network_path + ':/home/pi/' + source_filename
    callstr = 'scp ' + qflag + src + ' ' + target_filename
    _ = os.system(callstr)

def read_files_from_RPi(files, target_path, address, username='pi', verbose=False):
    """
    files - list or tuple of files to copy from home directory of the RPi
    target_path - folder to which the files will be copied to
    address - RPi address
    username - RPi username to use for access
    """
    for file in files:
        target_filename = os.path.join(target_path, file)
        read_file_from_RPi(file, target_filename, address, username, verbose)


class Camera_RPi_file_manager(object):
    def __init__(self, address, username='pi', verbose=False):
        self.address = address
        self.username = username
        self.verbose = verbose

    @staticmethod
    def Camera_RPi_files():
        return (os.path.join(package_path, 'ZMQcomms.py'),
                os.path.join(package_path, 'CameraRPiController.py'))

    @staticmethod
    def video_file_name_on_RPi():
        return 'video.h264'

    @staticmethod
    def video_file_name_on_RecordingPC(cameraID):
        return 'video_' + cameraID + '.h264'

    def update_files_on_RPi(self):
        write_files_to_RPi(Camera_RPi_file_manager.Camera_RPi_files(), self.address, self.username, self.verbose)

    def retrieve_timestamps(self):
        temp_folder = mkdtemp('RPiTempFolder')
        files = ('RawVideoEncoderTimestamps.csv', 
                 'VideoEncoderTimestamps.csv', 
                 'TTLpulseTimestamps.csv')
        read_files_from_RPi(files, temp_folder, self.address, self.username, self.verbose)
        filename = os.path.join(temp_folder, files[0])
        self.OnlineTrackerData_timestamps = np.genfromtxt(filename, dtype=int)
        filename = os.path.join(temp_folder, files[1])
        self.VideoData_timestamps = np.genfromtxt(filename, dtype=int)
        filename = os.path.join(temp_folder, files[2])
        self.GlobalClock_timestamps = np.genfromtxt(filename, dtype=int)
        rmtree(temp_folder)

    def retrieve_OnlineTrackerData(self):
        """
        Retrieves OnlineTrackerData. The data transfer failes sometimes, 
        in which case an informing message is printed and retrieval is attempted again.

        Note! OnlineTrackerData_timestamps must be obtained first using retrieve_timestamps() method. 
        """
        if hasattr(self, 'OnlineTrackerData_timestamps'):
            temp_folder = mkdtemp('RPiTempFolder')
            self.OnlineTrackerData = np.zeros((0, 0))
            while self.OnlineTrackerData.shape[0] != self.OnlineTrackerData_timestamps.size:
                try:
                    read_files_from_RPi(('OnlineTrackerData.csv',), temp_folder, self.address, self.username, self.verbose)
                    filename = os.path.join(temp_folder, 'OnlineTrackerData.csv')
                    self.OnlineTrackerData = np.genfromtxt(filename, delimiter=',', dtype=float)
                except ValueError:
                    print('Failed to get OnlineTrackerData, trying again at: ' + self.address)
            rmtree(temp_folder)
        else:
            raise Exception('OnlineTrackerData_timestamps must be obtained first.')

    def retrieve_timestamps_and_OnlineTrackerData(self):
        """
        Must be called before get_timestamps_and_OnlineTrackerData()
        """
        self.retrieve_timestamps()
        self.retrieve_OnlineTrackerData()

    def get_timestamps_and_OnlineTrackerData(self, cameraID='0'):
        """
        Must be called after retrieve_timestamps_and_OnlineTrackerData()
        """
        return {'ColumnLabels': ['X1', 'Y1', 'X2', 'Y2', 'Luminance_1', 'Luminance_2'], 
                'OnlineTrackerData': self.OnlineTrackerData,
                'OnlineTrackerData_timestamps': self.OnlineTrackerData_timestamps,
                'VideoData_timestamps': self.VideoData_timestamps,
                'GlobalClock_timestamps': self.GlobalClock_timestamps, 
                'VideoFile': Camera_RPi_file_manager.video_file_name_on_RecordingPC(cameraID)}

    def copy_over_video_data(self, folder_path, cameraID='0'):
        """
        Copies over video data to folder_path with cameraID included in the filename.
        """
        read_file_from_RPi(Camera_RPi_file_manager.video_file_name_on_RPi(), 
                           os.path.join(folder_path, Camera_RPi_file_manager.video_file_name_on_RecordingPC(cameraID)), 
                           self.address, self.username, self.verbose)


class CameraSimulation(object):

    def __init__(self, address, port):
        self.zmq_publisher = None
        self.init_zmq_publisher(address, port)
        self.position_messaging_thread = Thread(target=self.position_messaging_loop)
        self.keep_messaging = False

    def init_zmq_publisher(self, address, port):
        """
        Sets up publishing messages with ZMQ at 'localhost'.
        """
        # For sending position data
        contextpub = zmq.Context()
        self.zmq_publisher = contextpub.socket(zmq.PUB)
        self.zmq_publisher.bind('tcp://' + '*' + ':' + str(port))
        sleep(0.5)  # Give time to establish sockets for ZeroMQ

    def send_linedata(self, linedata):
        message = json.dumps(linedata)  # Convert data to string format
        message = message.encode()  # Convert data into bytes format
        self.zmq_publisher.send(message)  # Send the message using ZeroMQ

    def position_messaging_loop(self, interval=0.033):
        while self.keep_messaging:
            simulated_position = np.random.rand(2) * 50
            self.send_linedata([simulated_position[0], simulated_position[1], None, None, 100, None])
            sleep(interval)

    def sendCommand(self, command, return_output=False):
        if command in ('start_streaming', 'start_recording_video'):
            return
        ret = getattr(self, command)()
        if return_output:
            return ret

    def start(self):
        self.keep_messaging = True
        self.position_messaging_thread.start()

    def stop(self):
        self.keep_messaging = False
        self.position_messaging_thread.join()

    def start_processing(self):
        self.start()

    def close(self):
        self.keep_messaging = False
        self.position_messaging_thread.join()
        self.zmq_publisher.close()


class CameraControl(object):
    def __init__(self, address, port, username='pi', password='raspberry', 
                 resolution_option=None, OnlineTrackerParams=None, framerate=30, simulation=False):

        # This class skips some methods and uses CameraSimulation class instead of RemoteControl class
        # if simulation is set True.
        self.simulation = simulation

        # Make sure files on RPi are up to date
        if not simulation:
            self.RPi_file_manager = Camera_RPi_file_manager(address, username)
            self.RPi_file_manager.update_files_on_RPi()
        # Start initiation protocol and repeat if failed
        init_successful = False
        attempts = 0
        while not init_successful:
            if attempts > 4:
                reboot = True
                attempts = 0
            else:
                reboot = False
            init_successful = self._init_CameraController(address, port, username, password, 
                                                          resolution_option, OnlineTrackerParams, 
                                                          framerate, reboot)
            if not init_successful:
                attempts += 1
                print('Remote CameraController initialization failed, trying again at: ' + address)

    def _init_CameraController(self, address, port, username, password, 
                               resolution_option, OnlineTrackerParams, framerate, reboot=False):

        if self.simulation:
            address = '*' if address == 'localhost' else address
            self.RemoteControl = CameraSimulation(address, OnlineTrackerParams['OnlineTracker_port'])
            return True

        if hasattr(self, 'RemoteControl'):
            self.RemoteControl.close()
        self.RemoteControl = remote_object_controller(address, int(port))
        self.RPiSSH = self._init_RPiSSH(address, port, username, password, reboot)
        init_successful = self.RemoteControl.pair(10)
        if init_successful:
            init_successful = self.RemoteControl.sendInitCommand(10, 
                                                                 resolution_option, 
                                                                 framerate,
                                                                 OnlineTrackerParams)

        return init_successful

    def _init_RPiSSH(self, address, port, username='pi', password='raspberry', reboot=False):
        if not hasattr(self, 'RPiSSH') or self.RPiSSH is None:
            self.RPiSSH = ssh(address, username, password)
        if reboot:
            print('Rebooting at ' + address)
            self.RPiSSH.sendCommand('sudo reboot')
            self.RPiSSH.disconnect()
            self.RPiSSH = None
            sleep(30) # Assumed minimum rebooting time for Raspberry Pi
            pinging_successful = False
            while not pinging_successful:
                pinging_successful = test_pinging_address(address)
            self._init_RPiSSH(address, port, username, password, reboot=False)
        else:
            self.RPiSSH.sendCommand('sudo pkill python') # Ensure any past processes have closed
            command = 'source /home/pi/.virtualenvs/python3/bin/activate && '
            command += 'python CameraRPiController.py --remote --port ' + str(port)
            self.RPiSSH.sendCommand_threading(command) # TO DO: This seems to kill the ssh instance if command fails.

    def calibrate(self, calibration_parameters):
        """
        calibration_parameters - dict - output from CameraControl.enter_calibration_parameters
        """
        return self.RemoteControl.sendCommand('calibrate', True, calibration_parameters)

    def start_streaming(self, address, port):
        self.RemoteControl.sendCommand('start_streaming', True, address, int(port))

    def start_processing(self):
        self.RemoteControl.sendCommand('start_processing', True)

    def start(self):
        self.RemoteControl.sendCommand('start_recording_video', True)
        self.RemoteControl.sendCommand('start_processing', True)

    def stop(self):
        self.RemoteControl.sendCommand('stop', True)

    def close(self):
        self.RemoteControl.sendCommand('close', True)
        self.RemoteControl.close()
        if hasattr(self, 'RPiSSH') and not (self.RPiSSH is None):
            self.RPiSSH.disconnect()

    @staticmethod
    def dict_OnlineTrackerParams(calibrationTmatrix, tracking_mode, smoothing_box, 
                                 motion_threshold, motion_size, 
                                 OnlineTracker_port, RPiIP, LED_separation):
        """
        Returns camera_parameters in a dictionary.
        """
        return {'calibrationTmatrix': calibrationTmatrix, # (3 x 3 numpy array)
                'tracking_mode': str(tracking_mode), # str
                'smoothing_box': int(smoothing_box), # int - number of pixels
                'motion_threshold': motion_threshold,  # int
                'motion_size': motion_size,  # int
                'OnlineTracker_port': str(OnlineTracker_port), # str
                'RPiIP': str(RPiIP), # str
                'LED_separation': float(LED_separation)} # LED separation in cm

    @staticmethod
    def dict_calibration_parameters(ndots_xy, spacing, offset_xy):
        """
        Returns calibration_parameters in a dictionary.
        """
        return {'ndots_xy': ndots_xy, 
                'spacing': spacing, 
                'offset_xy': offset_xy}

    @staticmethod
    def CameraControl_unpackPool(kwargs):
        """
        Helper function to allow using pool with argument list.
        kwargs - dict - key value pairs for input arguments to CameraControl
        """
        return CameraControl(**kwargs)

    @staticmethod
    def init_CameraControls_with_CameraSettings(CameraSettings, simulation=False):
        """
        Returns a dictionary where each CameraSettings['use_RPi_nrs'] element is a key
        to corresponding initialized CameraControl object. 
        """
        kwargs_list = []
        cameraID_list = []
        for cameraID in CameraSettings['CameraSpecific'].keys():
            cameraID_list.append(cameraID)
            args = ((CameraSettings['CameraSpecific'][cameraID]['CalibrationData']['low']['calibrationTmatrix']
                     if not simulation else np.ones((3, 3))),
                    CameraSettings['General']['tracking_mode'], 
                    CameraSettings['General']['smoothing_box'], 
                    CameraSettings['General']['motion_threshold'], 
                    CameraSettings['General']['motion_size'], 
                    CameraSettings['General']['OnlineTracker_port'], 
                    CameraSettings['CameraSpecific'][cameraID]['address'], 
                    CameraSettings['General']['LED_separation'])
            OnlineTrackerParams = CameraControl.dict_OnlineTrackerParams(*args)
            kwargs_list.append({'address': CameraSettings['CameraSpecific'][cameraID]['address'],
                                'port': CameraSettings['General']['ZMQcomms_port'],
                                'username': CameraSettings['General']['username'],
                                'password': CameraSettings['General']['password'],
                                'resolution_option': CameraSettings['General']['resolution_option'],
                                'framerate': CameraSettings['General']['framerate'],
                                'OnlineTrackerParams': OnlineTrackerParams,
                                'simulation': simulation
                                })
        pool = ThreadPool(len(kwargs_list))
        CameraControlsList = pool.map(CameraControl.CameraControl_unpackPool, kwargs_list)
        pool.close()
        CameraControls = {}
        for camera_controller, cameraID in zip(CameraControlsList, cameraID_list):
            CameraControls[cameraID] = camera_controller

        return CameraControls


class TrackingControl(object):

    def __init__(self, CameraSettings, simulation=False):
        self.CameraControls = CameraControl.init_CameraControls_with_CameraSettings(CameraSettings,
                                                                                    simulation=simulation)

    def start(self):
        T_list = []
        for cameraID in self.CameraControls.keys():
            T = Thread(target=self.CameraControls[cameraID].start)
            T.start()
        for T in T_list:
            T.join()

    def stop(self):
        T_list = []
        for cameraID in self.CameraControls.keys():
            T = Thread(target=self.CameraControls[cameraID].stop)
            T.start()
        for T in T_list:
            T.join()

    def close(self):
        T_list = []
        for cameraID in self.CameraControls.keys():
            T = Thread(target=self.CameraControls[cameraID].close)
            T.start()
        for T in T_list:
            T.join()


class onlineTrackingData(object):
    # Constantly updates position data for all RPis currently in use.
    # Initialize this class as RPIpos = onlineTrackingData(CameraSettings, arena_size)
    # Check for latest position with combPos = RPIpos.combPosHistory[-1]

    # RPIpos.combPosHistory is multiprocessing.managers.List and can take care of locks across processes

    # Optional arguments during initialization:
    #   HistogramParameters is a list: [margins, binSize, histogram_speed_limit]
    def __init__(self, CameraSettings, arena_size, HistogramParameters=None):
        # Initialise the class with input CameraSettings
        self.combPos_update_interval = 0.05 # in seconds
        self.CameraSettings = CameraSettings
        self.arena_size = arena_size
        # Make a list of cameraIDs from CameraSettings for reference to other lists in this class.
        self.cameraIDs = sorted(self.CameraSettings['CameraSpecific'].keys())
        if HistogramParameters is None:
            HistogramParameters = {'margins': 10,  # histogram data margins in centimeters
                                   'binSize': 2,  # histogram binSize in centimeters
                                   'speedLimit': 10}  # centimeters of distance in last second to be included
        self.KeepGettingData = True # Set True for endless while loop of updating latest data
        self.posDatas = [None for i in range(len(self.cameraIDs))]
        self.multiprocess_manager = multiprocessing.Manager()
        self.combPosHistory = self.multiprocess_manager.list([])
        self.sockSUBs = onlineTrackingData.setupSockets(self.cameraIDs, self.CameraSettings)
        # Initialize Locks to avoid errors
        self.posDatasLock = Lock()
        # Initialize histogram
        self.position_histogram_array = None
        self.position_histogram_dict = None
        self.initialize_histogram(**HistogramParameters)
        # Start updating position data and storing it in history
        self.T_updateCombPosHistory = Thread(target=self.updateCombPosHistory)
        self.T_updateCombPosHistory.start()

        self.is_alive = self.multiprocess_manager.Value(bool, True)

    @staticmethod
    def setupSocket(address, port):
        # Set ZeroMQ socket to listen on incoming position data from all RPis
        context = zmq.Context()
        sockSUB = context.socket(zmq.SUB)
        sockSUB.setsockopt(zmq.SUBSCRIBE, ''.encode())
        sockSUB.RCVTIMEO = 100  # maximum duration to wait for data (in milliseconds)
        sockSUB.connect('tcp://{}:{}'.format(address, port))

        return sockSUB

    @staticmethod
    def setupSocket_unpackPool(kwargs):
        """
        Helper function to allow using pool with argument list.
        kwargs - dict - key value pairs for input arguments to setupSocket
        """
        return onlineTrackingData.setupSocket(**kwargs)

    @staticmethod
    def setupSockets(cameraIDs, CameraSettings):
        kwargs_list = []
        for cameraID in cameraIDs:
            kwargs_list.append({'address': CameraSettings['CameraSpecific'][cameraID]['address'],
                                'port': CameraSettings['General']['OnlineTracker_port']})
        pool = ThreadPool(len(cameraIDs))
        sockSUBs = pool.map(onlineTrackingData.setupSocket_unpackPool, kwargs_list)
        pool.close()

        return sockSUBs

    def updatePosDatas(self, nRPi):
        # Updates self.posDatas when any new position data is received
        # This loop continues until self.KeepGettingData is set False. This is done by self.close function
        while self.KeepGettingData:
            # Wait for position data update
            try:
                message = self.sockSUBs[nRPi].recv()  # Receive message
                message = message.decode()  # Decode bytes into string
            except:
                message = 'no message'
            if message != 'no message':
                posData = json.loads(message)  # Convert from string to original format
                # Ignore messages where all elements are None
                if any(posData):
                    # Update posData for the correct position in the list
                    with self.posDatasLock:
                        self.posDatas[nRPi] = posData

    def combineCurrentLineData(self, previousCombPos):
        with self.posDatasLock:
            posDatas = self.posDatas
        # Combine posDatas from cameras to position data
        if len(posDatas) > 1:
            # Convert posDatas for use in combineCamerasData function
            cameraPos = []
            for posData in posDatas:
                cameraPos.append(np.array(posData[:4], dtype=np.float32))
            # Combine data from cameras
            lastCombPos = combineCamerasData(cameraPos, previousCombPos, self.cameraIDs, 
                                             self.CameraSettings, self.arena_size)
        else:
            # If only a single camera is used, extract position data from posData into numpy array
            lastCombPos = np.array(posDatas[0][:4], dtype=np.float32)

        return lastCombPos

    def initialize_histogram(self, margins, binSize, speedLimit):
        # Initialise histogram edgesrameters
        xHistogram_edges = np.append(np.arange(-margins, self.arena_size[0] + margins, binSize), 
                                     self.arena_size[0] + margins)
        yHistogram_edges = np.append(np.arange(-margins, self.arena_size[1] + margins, binSize), 
                                     self.arena_size[1] + margins)
        histmap = np.zeros((yHistogram_edges.size - 1, xHistogram_edges.size - 1), dtype=np.int16)
        # Create shared data
        self.position_histogram_array = multiprocessing.Array('I', histmap.flatten(), lock=True)
        self.position_histogram_dict = {
            'array_shape': histmap.shape,
            'parameters': {'margins': margins,
                           'binSize': binSize,
                           'speedLimit': speedLimit},
            'edges': {'x': xHistogram_edges, 'y': yHistogram_edges}
        }

    def updateCombPosHistory(self):
        # Initialize RPi position data listening, unless synthetic data requested
        self.T_updatePosDatas = []
        for nRPi in range(len(self.posDatas)):
            T = Thread(target=self.updatePosDatas, args=(nRPi,))
            T.start()
            self.T_updatePosDatas.append(T)
        # Continue once data is received from each RPi
        RPi_data_available = np.zeros(len(self.cameraIDs), dtype=bool)
        while not np.all(RPi_data_available) and self.KeepGettingData:
            sleep(0.05)
            for nRPi in range(len(self.posDatas)):
                if not (self.posDatas[nRPi] is None):
                    RPi_data_available[nRPi] = True
        if not self.KeepGettingData:
            return
        print('All RPi data available')
        # Set up speed tracking
        one_second_steps = int(np.round(1 / self.combPos_update_interval))
        self.lastSecondDistance = 0 # vector distance from position 1 second in past
        # Check data is available before proceeding
        lastCombPos = None
        while lastCombPos is None and self.KeepGettingData:
            lastCombPos = self.combineCurrentLineData(None)
        if not self.KeepGettingData:
            return
        self.combPosHistory.append(list(lastCombPos))
        time_of_last_datapoint = time()
        # Update the data at specific interval
        while self.KeepGettingData:
            time_since_last_datapoint = time() - time_of_last_datapoint
            if time_since_last_datapoint > self.combPos_update_interval:

                # If enough time has passed since last update, append to combPosHistory list

                lastCombPos = self.combineCurrentLineData(self.combPosHistory[-1])

                if not (lastCombPos is None):
                    self.combPosHistory.append(list(lastCombPos))
                else:
                    self.combPosHistory.append(lastCombPos)

                time_of_last_datapoint = time()
                if len(self.combPosHistory) > one_second_steps:
                    # Compute distance from one second in the past if enough data available
                    tmp_combPosHistory_segment = np.array(self.combPosHistory[-one_second_steps:])
                    currPos = tmp_combPosHistory_segment[-1]
                    pastPos = tmp_combPosHistory_segment[0]
                    if not (currPos is None) and not (pastPos is None):
                        self.lastSecondDistance = euclidean(currPos[:2], pastPos[:2])
                        if self.lastSecondDistance > self.position_histogram_dict['parameters']['speedLimit']:

                            # If animal has been moving enough, update histogram
                            position_histogram_indices = (
                                np.argmin(np.abs(self.position_histogram_dict['edges']['y'][:-1] - currPos[1])),
                                np.argmin(np.abs(self.position_histogram_dict['edges']['x'][:-1] - currPos[0]))
                            )
                            ind = np.ravel_multi_index(position_histogram_indices,
                                                       self.position_histogram_dict['array_shape'])

                            self.position_histogram_array[ind] += 1

                    else:
                        self.lastSecondDistance = None
            else:
                sleep(self.combPos_update_interval * 0.1)

        self.is_alive.set(False)

    def close(self):
        # Closes the updatePosDatas thread and ZeroMQ socket for position listening
        self.KeepGettingData = False
        for T in self.T_updatePosDatas:
            T.join()
        self.T_updateCombPosHistory.join()
        for sockSUB in self.sockSUBs:
            sockSUB.close()


class RewardControl(object):
    # This class allows control of FEEDERs
    # FEEDER_type can be either 'milk' or 'pellet'
    def __init__(self, FEEDER_type, RPiIP, RPiUsername, RPiPassword, trialAudioSignal=None, 
                 negativeAudioSignal=0, lightSignalIntensity=0, lightSignalPins=[]):
        self.FEEDER_type = FEEDER_type
        self.RPiIP = RPiIP
        self.trialAudioSignal = trialAudioSignal
        self.lightSignalIntensity = lightSignalIntensity
        self.negativeAudioSignal = negativeAudioSignal
        self.lightSignalPins = lightSignalPins
        # Ensure files on the FEEDER RPi are up to date
        RewardControl.update_files_on_RPi(FEEDER_type, RPiIP, RPiUsername, verbose=False)
        # Set up SSH connection
        self.ssh_connection = ssh(RPiIP, RPiUsername, RPiPassword)
        self.ssh_connection.sendCommand('sudo pkill python') # Ensure any past processes have closed
        self.ssh_Controller_connection = ssh(RPiIP, RPiUsername, RPiPassword)
        # Set up ZMQ connection
        self.Controller_messenger = paired_messenger(address=self.RPiIP, port=4186)
        self.Controller_messenger.add_callback(self.Controller_message_parser)
        # Initialize Controller
        init_successful = self.init_Controller_until_positive_feedback(max_attempts=4, max_init_wait_time=25)
        # If initialization was unsuccessful, crash this script
        if not init_successful:
            raise Exception('Initialization was unsuccessful at: ' + RPiIP)

    @staticmethod
    def update_files_on_RPi(FEEDER_type, address, username, verbose=False):
        if FEEDER_type == 'milk':
            files = (os.path.join(package_path, 'milkFeederController.py'),)
        elif FEEDER_type == 'pellet':
            files = (os.path.join(package_path, 'pelletFeederController.py'),)
        else:
            raise ValueError('Unexpected FEEDER_type argument.')
        write_files_to_RPi(files, address, username=username, verbose=verbose)

    def Controller_message_parser(self, message):
        if message == 'init_successful'.encode():
            self.Controller_init_successful = True
        if message == 'releaseMilk successful'.encode():
            self.release_feedback_message_received = True
            self.releaseMilk_successful = True
        if message == 'release_pellets successful'.encode():
            self.release_feedback_message_received = True
            self.release_pellets_successful = True
        if message == 'release_pellets failed'.encode():
            self.release_feedback_message_received = True
            self.release_pellets_successful = False
        if message == 'release_pellets in_progress'.encode():
            self.release_pellets_in_progress = True

    def milkFeederController_init_command(self):
        command = 'python milkFeederController.py'
        command += ' ' + '--pinchValve '
        command += ' ' + '--init_feedback'
        if not (self.trialAudioSignal is None):
            command += ' ' + '--trialAudioSignal ' + ' '.join(map(str, self.trialAudioSignal))
        if self.negativeAudioSignal > 0:
            command += ' ' + '--negativeAudioSignal ' + str(self.negativeAudioSignal)
        if self.lightSignalIntensity > 0:
            command += ' ' + '--lightSignalIntensity ' + str(self.lightSignalIntensity)
        if len(self.lightSignalPins) > 0:
            command += ' ' + '--lightSignalPins ' + ' '.join(map(str, self.lightSignalPins))

        return command

    def pelletFeederController_init_command(self):
        command = 'python pelletFeederController.py'
        command += ' ' + '--init_feedback'

        return command

    def init_Controller_and_wait_for_feedback(self, max_init_wait_time=25):
        """
        Attempts to initialize the Controller class on the FEEDER
        """
        self.Controller_init_successful = False
        # Acquire the correct command for the FEEDER type
        if self.FEEDER_type == 'milk':
            command = self.milkFeederController_init_command()
        elif self.FEEDER_type == 'pellet':
            command = self.pelletFeederController_init_command()
        else:
            raise Exception('Unknown FEEDER_type {}'.format(self.FEEDER_type))
        # Initiate process on the RPi over SSH
        # Set timeout to max_init_wait_time and verbosity to False
        self.ssh_Controller_connection.sendCommand(command, max_init_wait_time, False)
        # Wait until positive initialization feedback or timer runs out
        start_time = time()
        while not self.Controller_init_successful and (time() - start_time) < max_init_wait_time:
            sleep(0.1)

        return self.Controller_init_successful

    def init_Controller_until_positive_feedback(self, max_attempts=4, max_init_wait_time=25):
        """
        Attempts to initialize Controller class on the FEEDER multiple times
        """
        controller_init_successful = False
        for n in range(max_attempts):
            controller_init_successful = \
                self.init_Controller_and_wait_for_feedback(max_init_wait_time=max_init_wait_time)
            if controller_init_successful:
                break
            else:
                print('Controller activation failed at ' + self.RPiIP + ' Re-initializing ...')
                # If initiation is unsuccessful, start again after killing existing processes
                self.ssh_connection.sendCommand('sudo pkill python') # Ensure any past processes have closed

        return controller_init_successful

    def release(self, quantity=1, max_attempts=10):
        """
        Releases the specified quantity of reward.
        Returns True or False, depending if action was successful.
        """
        # Reset feedback message detector
        self.release_feedback_message_received = False
        self.release_pellets_in_progress = False
        # Compute maximum time to wait for feedback depending on FEEDER type
        if self.FEEDER_type == 'milk':
            max_wait_time = quantity + 4
        elif self.FEEDER_type == 'pellet':
            max_wait_time = 5 * max_attempts
        else:
            raise Exception('Unknown FEEDER_type {}'.format(self.FEEDER_type))
        # Send ZMQ command depending on FEEDER type
        if self.FEEDER_type == 'milk':
            command = 'releaseMilk ' + str(quantity) + ' True'
        elif self.FEEDER_type == 'pellet':
            command = 'release_pellets ' + str(quantity) + ' True'
        else:
            raise Exception('Unknown FEEDER_type {}'.format(self.FEEDER_type))
        self.Controller_messenger.sendMessage(command.encode())
        # Wait for feedback message
        start_time = time()
        while not self.release_feedback_message_received and (time() - start_time) < max_wait_time:
            sleep(0.1)
            # The following allows reseting start_time between pellets
            if self.release_pellets_in_progress:
                self.release_pellets_in_progress = False
                start_time = time()
        # Return True or False if message successful/failure or False if no message received in time.
        if self.release_feedback_message_received:
            if self.FEEDER_type == 'milk':
                return self.releaseMilk_successful
            elif self.FEEDER_type == 'pellet':
                return self.release_pellets_successful
        else:
            return False

    def startTrialAudioSignal(self):
        self.Controller_messenger.sendMessage('startTrialAudioSignal'.encode())

    def stopTrialAudioSignal(self):
        self.Controller_messenger.sendMessage('stopTrialAudioSignal'.encode())

    def playNegativeAudioSignal(self):
        self.Controller_messenger.sendMessage('playNegativeAudioSignal'.encode())

    def startLightSignal(self):
        self.Controller_messenger.sendMessage('startLightSignal'.encode())

    def stopLightSignal(self):
        self.Controller_messenger.sendMessage('stopLightSignal'.encode())

    def startAllSignals(self):
        self.Controller_messenger.sendMessage('startAllSignals'.encode())

    def stopAllSignals(self):
        self.Controller_messenger.sendMessage('stopAllSignals'.encode())

    def close(self):
        """
        Closes all opened processes correctly.
        """
        try:
            self.Controller_messenger.sendMessage('close'.encode())
            self.Controller_messenger.close()
            self.ssh_Controller_connection.disconnect()
            self.ssh_connection.disconnect()
        except Exception as e:
            from inspect import currentframe, getframeinfo
            frameinfo = getframeinfo(currentframe())
            print('Error in ' + frameinfo.filename + ' line ' + str(frameinfo.lineno - 3))
            print('Failed to close connection: ' + str(self.RPiIP))
            print(e)

class GlobalClockControl(object):

    def __init__(self, address, port, username='pi', password='raspberry'):
        self.initController_messenger(address, port)
        self.T_initRPiController = Thread(target=self.initRPiController, 
                                          args=(address, port, username, password))
        self.T_initRPiController.start()
        # Wait until all RPis have confirmed to be ready
        self.RPi_init_Successful = False
        while not self.RPi_init_Successful:
            sleep(0.1)

    def initController_messenger(self, address, port):
        # Set up ZMQ connection
        self.Controller_messenger = paired_messenger(address=address, port=int(port))
        self.Controller_messenger.add_callback(self.Controller_message_parser)
        sleep(1)

    def initRPiController(self, address, port, username, password):
        self.RPiSSH = ssh(address, username, password)
        self.RPiSSH.sendCommand('sudo pkill python')  # Ensure any past processes have closed
        command = 'python GlobalClock.py --remote --port ' + str(port)
        self.RPiSSH.sendCommand(command)

    def Controller_message_parser(self, message):
        if message == 'init_successful'.encode():
            self.RPi_init_Successful = True

    def start(self):
        self.Controller_messenger.sendMessage('start'.encode())

    def stop(self):
        self.Controller_messenger.sendMessage('stop'.encode())

    def close(self):
        self.stop()
        self.Controller_messenger.sendMessage('close'.encode())
        # Close SSH connections
        self.RPiSSH.disconnect()
        self.T_initRPiController.join()
        # Close ZMQ messengers
        self.Controller_messenger.close()



from RPi import GPIO
from time import time, sleep
from copy import copy
from ZMQcomms import paired_messenger
import argparse

GPIO.setmode(GPIO.BOARD) # Use board numbering for TTL pin

class detect_pellet(object):
    '''
    This class signals when IR break beam has been interrupted.
    It can call a callback function and/or pause process until pellet detection.
    At pellet detection, the class closes automaticall, unless single_detection is set to False.
    '''

    def __init__(self, callback=None, single_detection=True, beamPin=15):
        '''
        callback - optional - function to call when IR break beam has been interrupted
        single_detection - bool - True by default, 
                           which closes detect_pellet class after first time IR beam is interrupted.
        beamPin - the GPIO.BOARD pin numbering
        '''
        self.callback = callback
        self.single_detection = single_detection
        self.beamPin = beamPin
        self.pellet_detected = False
        GPIO.setup(self.beamPin, GPIO.IN, pull_up_down=GPIO.PUD_UP)
        GPIO.add_event_detect(self.beamPin, GPIO.FALLING, callback=self.beam_interrupted)

    def beam_interrupted(self, channel):
        '''
        This method is called when IR break beam is interrupted.
        It closes the detect_pellet bindings, if single_dection=True.
        It changes the attribute pellet_detected to True.
        It calls the callback function in __init__ input arguments.
        '''
        if self.single_detection:
            self.close()
        self.pellet_detected = True
        if not (self.callback is None):
            self.callback()

    def wait_for_pellet_detection(self, max_wait_time=1.0):
        '''
        This method waits for pellet detection for the maximum time stated.
        It exits as soon as pellet is detected or timer runs out, closing detect_pellet class.
        It returns True or False, depending if pellet was detected.
        '''
        starting_time = time()
        while not self.pellet_detected and (time() - starting_time) < max_wait_time:
            sleep(0.05)
        self.close()

        return self.pellet_detected

    def close(self):
        GPIO.cleanup(self.beamPin)


class servo_controller(object):
    '''
    This class controls servo SG-5010 with PWM signal
    '''    
    def __init__(self, servoPin=12, startingAngle=90):
        self.servoPin = servoPin # Specify the TTL pin to be used for signal on RPi board
        self.PWM_cycle_frequency = 50.0 # in Hz
        self.time_for_180_turn = 1.0 # in seconds
        GPIO.setup(self.servoPin, GPIO.OUT)
        # Initialize PWM signalling
        self.pwm = GPIO.PWM(self.servoPin, self.PWM_cycle_frequency)
        GPIO.output(self.servoPin, True)
        # Set motor to starting angle position
        self.current_angle = startingAngle
        self.pwm.start(self.compute_duty_cycle_for_angle(self.current_angle))
        sleep(self.time_for_180_turn / 2.0)

    def compute_duty_cycle_for_angle(self, angle):
        return angle / 19.5 + 2.0

    def setAngle(self, angle):
        '''
        Instructs the stepper motor to move to specific angle by changing the duty cycle of the PWM signal.
        '''
        self.pwm.ChangeDutyCycle(self.compute_duty_cycle_for_angle(angle))
        self.current_angle = angle

    def computeMoveTimeToAngle(self, angle):
        '''
        Returns an estimated time of seconds required to turn to given angle from current angle.
        '''
        angular_difference = abs(self.current_angle - angle)
        proportion_of_180_turn = float(angular_difference) / 180.0

        return proportion_of_180_turn * self.time_for_180_turn

    def moveToAngle(self, angle):
        '''
        Moves stepper motor to specified angle.
        '''
        time_required = self.computeMoveTimeToAngle(angle)
        self.setAngle(angle)
        sleep(time_required)

    def wiggle(self, angular_range=15, n_repeats=1):
        '''
        Wiggles stepper motor back and forth the amount of degrees specified in angular_range and n_repeats times.
        '''
        original_angle = copy(self.current_angle)
        time_required = self.computeMoveTimeToAngle(original_angle + angular_range)
        for n in range(n_repeats):
            self.setAngle(original_angle + angular_range)
            sleep(time_required)
            self.setAngle(original_angle - angular_range)
            sleep(2 * time_required)
            self.setAngle(original_angle)
            sleep(time_required)
    
    def close(self):
        self.pwm.stop()
        sleep(0.1) # Ensures PWM stops properly before signal is terminated.
        GPIO.output(self.servoPin, False)
        GPIO.cleanup(self.servoPin)


def release_pellet(LoadingAngle=30, ReleaseAngle=90, wiggle_angular_range=15, wiggle_repeats=4):
    '''
    This function attempts to release a single pellet with stepper motor
    '''
    sc = servo_controller()
    sc.moveToAngle(LoadingAngle)
    sc.wiggle(wiggle_angular_range, wiggle_repeats)
    sc.moveToAngle(ReleaseAngle)
    sc.close()


def release_exactly_one_pellet(LoadingAngle=30, ReleaseAngle=90, wiggle_angular_range=15, wiggle_repeats=4, max_attempts=10):
    '''
    This function attempts to release a pellet until one is detected by the IR break beam.
    It waits until completed and returns either True or False depending if successful.
    '''

    max_wait_time_for_pellet=1.0
    for n in range(max_attempts):
        dp = detect_pellet()
        release_pellet(LoadingAngle, ReleaseAngle, wiggle_angular_range, wiggle_repeats)
        pellet_released = dp.wait_for_pellet_detection(max_wait_time=max_wait_time_for_pellet)
        if pellet_released:
            break

    return pellet_released


class Controller(object):
    '''
    This class controls the operation of the stepper motor and IR break beam 
    for releasing controlled number of pellets.
    '''

    def __init__(self, init_feedback=False):
        '''
        init_feedback - bool - Sends ZMQ message to localhost if initialization successful: init_successful
                        When using this, ensure a ZMQ device is paired to the IP of the device running this script.
        '''
        # Initialize ZMQcomms
        self.initialize_ZMQcomms()
        # Send feedback text if requested
        if init_feedback:
            self.ZMQmessenger.sendMessage('init_successful')

    def initialize_ZMQcomms(self):
        self.ZMQmessenger = paired_messenger(port=4186)
        self.ZMQmessenger.add_callback(self.command_parser)
        sleep(1) # This ensures all ZMQ protocols have been properly initated before finishing this process

    def command_parser(self, message):
        '''
        Parses incoming ZMQ message for function name, input arguments and calls that function.
        '''
        method_name = message.split(' ')[0]
        args = message.split(' ')[1:]
        getattr(self, method_name)(*args)

    def release_pellets(self, quantity=1, feedback=False):
        '''
        Releases the number of pellets requested. If feedback requested, sends bool in str format after each pellet.
        If an attempt of a single pellet fails in max_attempts, the rest are aborted.
        '''
        # Parse input arguments, converting from string format if necessary
        if type(quantity) != int:
            quantity = int(quantity)
        if type(feedback) != bool:
            if type(feedback) == str and feedback == 'True':
                feedback = True
            else:
                feedback = False
        # Attempt releasing all required pellets
        for n_pellet in range(quantity):
            self.ZMQmessenger.sendMessage('release_pellets in_progress')
            successful_release = release_exactly_one_pellet(LoadingAngle=30, ReleaseAngle=90, 
                                                            wiggle_angular_range=15, wiggle_repeats=4, 
                                                            max_attempts=10)
            # If release has failed, the rest of the pellets are aborted.
            if not successful_release:
                break
        # Return feedback over ZMQ if requested.
        if feedback:
            if successful_release:
                self.ZMQmessenger.sendMessage('release_pellets successful')
            else:
                self.ZMQmessenger.sendMessage('release_pellets failed')

    def keepProcessRunning(self):
        '''
        This method allows any process to wait until close method has been called for this class.
        '''
        self.closeCommandReceived = False
        while not self.closeCommandReceived:
            try:
                sleep(0.1)
            except KeyboardInterrupt:
                self.close()

    def close(self):
        '''
        Closes all initiated processes gracefully.
        '''
        self.closeCommandReceived = True
        self.ZMQmessenger.close()


if __name__ == '__main__':
    # Input argument handling and help info
    parser = argparse.ArgumentParser(description='Running this script initates Controller class.')
    parser.add_argument('--init_feedback', action='store_true', 
                        help='Sends ZMQ message to localhost if initialization successful: init_successful\n' + \
                        'When using this, ensure a ZMQ device is paired to the IP of the device running this script.')
    parser.add_argument('--releasePellet', type=int, nargs = 1, 
                        help='Makes device release a specified number of pellets and does not activate Controller class.')
    args = parser.parse_args()
    # If releasePellet command given skip everything and just release the number of pellets specified
    if args.releasePellet:
        for n_pellet in range(int(args.releasePellet[0])):
            successful_release = release_exactly_one_pellet(LoadingAngle=30, ReleaseAngle=90, 
                                                            wiggle_angular_range=15, wiggle_repeats=4, 
                                                            max_attempts=10)
            if not successful_release:
                break
    else:
        # Load input arguments
        if args.init_feedback:
            init_feedback = True
        else:
            init_feedback = False
        # Initialize Controller
        active_Controller = Controller(init_feedback=init_feedback)
        # Start an endless loop that can be cancelled with close method or ctrl+c
        active_Controller.keepProcessRunning()
